from ._utils import api_util
from .Bases import Iterable
from dss_python_backend.enums import ActionCodes

class ISwtControls(Iterable):
    __slots__ = []

    __name__ = "SwtControls"
    _api_prefix = "SwtControls"
    _columns = [
        "Name",
        "Idx",
        "Action",
        "Delay",
        "IsLocked",
        "NormalState",
        "State",
        "SwitchedObj",
        "SwitchedTerm",
    ]

    def Reset(self):
        self._lib.SwtControls_Reset()

    def Action(self, *args):
        """
        Open or Close the switch. No effect if switch is locked.  However, Reset removes any lock and then closes the switch (shelf state).

        Original COM help: https://opendss.epri.com/Action1.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.SwtControls_Get_Action()

        # Setter
        (Value,) = args
        self._lib.SwtControls_Set_Action(Value)

    def Delay(self, *args):
        """
        Time delay [s] between arming and opening or closing the switch.  Control may reset before actually operating the switch.

        Original COM help: https://opendss.epri.com/Delay3.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.SwtControls_Get_Delay()

        # Setter
        (Value,) = args
        self._lib.SwtControls_Set_Delay(Value)

    def IsLocked(self, *args):
        """
        The lock prevents both manual and automatic switch operation.

        Original COM help: https://opendss.epri.com/IsLocked.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.SwtControls_Get_IsLocked()

        # Setter
        (Value,) = args
        self._lib.SwtControls_Set_IsLocked(Value)

    def NormalState(self, *args):
        """
        Get/set the Normal state of switch (see ActionCodes) dssActionOpen or dssActionClose
        """
        # Getter
        if len(args) == 0:
            return ActionCodes(self._lib.SwtControls_Get_NormalState())

        # Setter
        (Value,) = args
        self._lib.SwtControls_Set_NormalState(Value)

    def State(self, *args):
        """
        Set it to force the switch to a specified state, otherwise read its present state.

        Original COM help: https://opendss.epri.com/State.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.SwtControls_Get_State()

        # Setter
        (Value,) = args
        self._lib.SwtControls_Set_State(Value)

    def SwitchedObj(self, *args):
        """
        Full name of the switched element.

        Original COM help: https://opendss.epri.com/SwitchedObj3.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.SwtControls_Get_SwitchedObj()

        # Setter
        (Value,) = args
        self._lib.SwtControls_Set_SwitchedObj(Value)

    def SwitchedTerm(self, *args):
        """
        Terminal number where the switch is located on the SwitchedObj

        Original COM help: https://opendss.epri.com/SwitchedTerm3.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.SwtControls_Get_SwitchedTerm()

        # Setter
        (Value,) = args
        self._lib.SwtControls_Set_SwitchedTerm(Value)


_SwtControls = ISwtControls(api_util)

# For backwards compatibility, bind to the default instance
Action = _SwtControls.Action
AllNames = _SwtControls.AllNames
Count = _SwtControls.Count
Delay = _SwtControls.Delay
First = _SwtControls.First
Idx = _SwtControls.Idx
IsLocked = _SwtControls.IsLocked
Name = _SwtControls.Name
Next = _SwtControls.Next
NormalState = _SwtControls.NormalState
Reset = _SwtControls.Reset
State = _SwtControls.State
SwitchedObj = _SwtControls.SwitchedObj
SwitchedTerm = _SwtControls.SwitchedTerm
_columns = _SwtControls._columns
__all__ = [
    "Action",
    "AllNames",
    "Count",
    "Delay",
    "First",
    "Idx",
    "IsLocked",
    "Name",
    "Next",
    "NormalState",
    "Reset",
    "State",
    "SwitchedObj",
    "SwitchedTerm",
]



# Import DSS-Python
import dss as dss_py
import os

# Integrate "Use environment variable for numpy version" from @kdheepak
# https://github.com/dss-extensions/OpenDSSDirect.py/pull/103/
# Changed for v0.10: defaults to True now.
OPENDSSDIRECT_PY_USE_NUMPY = os.environ.get("OPENDSSDIRECT_PY_USE_NUMPY", "1").upper() in ("1", "TRUE")

# Bind to the FFI module instance. This should be refined in a future version
lib = dss_py.prime_api_util.lib
ffi = dss_py.prime_api_util.ffi

api_util = dss_py.prime_api_util
DSSException = dss_py.DSSException



import inspect
import warnings
from dss_python_backend import enums
from dss import DSSException
from .Bases import Iterable

is_pandas_installed = True

try:
    import pandas as pd
except ImportError:
    is_pandas_installed = False


class Iterator:
    def __init__(self, module, function):
        warnings.warn("OpenDSSDirect.py's Iterator is deprecated; you can use native Python iterators directly now.", DeprecationWarning, stacklevel=2)
        assert inspect.ismodule(module) or isinstance(module, Iterable), "{module} must be of type module or a DSS iterable".format(
            module=module
        )
        self.module = module
        self.function = function

    def __iter__(self):
        from opendssdirect import dss

        try:
            idx = self.module.First()
        except:
            return # no circuit?

        if dss.Error.ExtendedErrors() and idx == 0:
                return # nothing to iterate

        while True:

            yield getattr(self.module, self.function)

            if not self.module.Next() > 0:
                break


def run_command(text, dss=None):
    """
    Use the Text interface of OpenDSS, grabbing all output text in a string.
    
    This is **deprecated** since it doesn't handle errors as exceptions and can confuse users.

    Prefer using [`dss("commands")`](#opendssdirect.OpenDSSDirect.OpenDSSDirect.__call__), [`dss.Text.Command("command")`](#opendssdirect.Text.IText.Command), or [`dss.Text.Commands("commands")`](#opendssdirect.Text.IText.Commands),.
    """

    warnings.warn('run_command is deprecated (use Command, Commands or the callable shortcut), see https://github.com/dss-extensions/OpenDSSDirect.py/issues/70', 
                  DeprecationWarning, stacklevel=2)

    if dss is None:
        from opendssdirect import dss

    api_util = dss.Basic._api_util
    r = []
    for l in text.splitlines():
        dss._lib.Text_Set_Command(l.encode(api_util.codec))
        r.append(api_util.get_string(dss._lib.Text_Get_Result()))

    return "\n".join(r).strip()


def try_impl(f):
    try:
        return f()
    except DSSException as ex:
        if ex.args[0] == 2:
            return None
        
        raise

def to_dataframe(module):
    """
    Iterate and export all data from the classic API class/module to a dataframe.

    If Pandas is not available, a dict of dicts is returned.
    """
    data = dict()

    for e in module:
        data[e.Name()] = dict()

    if len(data) != 0:
        for element in module:
            element_name = element.Name()
            data[element_name] = {
                n: try_impl(getattr(module, n)) for n, f in getmembers(module)
            }
    else:
        class_name = module.__name__
        warnings.warn("Empty list ({class_name})".format(class_name=class_name))

    if is_pandas_installed:
        return pd.DataFrame(data).T
    else:
        warnings.warn(
            "Pandas was not installed. Please see documentation for how to install extra dependencies."
        )
        return data


def _clean_data(data, class_name):
    from opendssdirect import dss

    for element in dss.ActiveClass.AllNames():
        name = "{class_name}.{element}".format(class_name=class_name, element=element)
        dss.ActiveClass.Name(element)

        all_prop_names = dss.Element.AllPropertyNames()
        if "NConds" in all_prop_names:
            nconds = int(data[name]["NConds"])
            x = []
            h = []
            units = []

            for cond in range(1, nconds + 1):
                dss("{name}.cond={cond}".format(name=name, cond=cond))

                dss("? {name}.x".format(name=name))
                x.append(float(dss.Text.Result()))

                dss("? {name}.h".format(name=name))
                h.append(float(dss.Text.Result()))

                dss("? {name}.units".format(name=name))
                units.append(dss.Text.Result())

            data[name]["X"] = x
            data[name]["H"] = h
            data[name]["Units"] = units

        elif "nconds" in all_prop_names: # backwards compat
            nconds = int(data[name]["nconds"])
            x = []
            h = []
            units = []

            for cond in range(1, nconds + 1):
                dss("{name}.cond={cond}".format(name=name, cond=cond))

                dss("? {name}.x".format(name=name))
                x.append(float(dss.Text.Result()))

                dss("? {name}.h".format(name=name))
                h.append(float(dss.Text.Result()))

                dss("? {name}.units".format(name=name))
                units.append(dss.Text.Result())

            data[name]["x"] = x
            data[name]["h"] = h
            data[name]["units"] = units

    return data


def class_to_dataframe(class_name, dss=None, transform_string=None, clean_data=None):
    """
    Export all DSS properties for a selected DSS object class.

    This function uses the `ActiveClass`, `Element` and `Properties` interfaces to iterate through the elements, extracting the value of each DSS property.

    Caveats are:

    - All properties are exported; this can be misleading since the order of definition and which properties are used can change how the DSS models behave.
    - All data is exported as text and must be reinterpreted in Python.
    - It can be slow for large circuits.

    Due to these caveats, the function has been marked **deprecated**. Still, it has been used by many third-party software as a way to extract important data, so it will be kept for backwards compatibility.

    For alternatives:
    
    - Check the functions for JSON output, such as [`Circuit.ToJSON()`](#opendssdirect.Circuit.ICircuit.ToJSON) (whole circuit), [`ActiveClass.ToJSON()`](#opendssdirect.ActiveClass.IActiveClass.ToJSON) (a whole class), and [`Element.ToJSON()`](#opendssdirect.Element.IElement.ToJSON). The DSS-Python examples are a good starting point (all related functions are also available in OpenDSSDirect.py): https://dss-extensions.org/DSS-Python/examples/JSON
    - The JSON output/input, and other future alternatives are based on the work-in-progress AltDSS Schema project. Check https://github.com/dss-extensions/AltDSS-Schema
    
    """
    warnings.warn("class_to_dataframe is deprecated; it will not be removed any time soon, using JSON exports is preferable. Watch AltDSS-Python for future alternatives, including native dataframes.", DeprecationWarning, stacklevel=2)

    if transform_string is None:
        transform_string = _evaluate_expression

    if clean_data is None:
        clean_data = _clean_data

    if not callable(transform_string):
        raise TypeError(
            "The `transform_string` must be a callable. Please check the documentation or contact the developer."
        )

    if dss is None:
        from opendssdirect import dss

    dss.Circuit.SetActiveClass("{class_name}".format(class_name=class_name))
    if class_name.lower() != dss.ActiveClass.ActiveClassName().lower():
        raise NotImplementedError(
            "`{class_name}` is not supported by the `class_to_dataframe` interface, please contact the developer for more information.".format(
                class_name=class_name
            )
        )
    data = dict()

    for element in dss.ActiveClass.AllNames():
        name = "{class_name}.{element}".format(class_name=class_name, element=element)
        dss.ActiveClass.Name(element)

        data[name] = dict()
        for i, n in enumerate(dss.Element.AllPropertyNames()):
            # use 1-based index for compatibility with previous versions
            string = dss.Properties.Value(str(i + 1))

            data[name][n] = transform_string(string)

    data = clean_data(data, class_name)

    if is_pandas_installed:
        return pd.DataFrame(data).T
    else:
        warnings.warn(
            "Pandas is not installed. Please see documentation for how to install extra dependencies."
        )
        return data


def _evaluate_expression(string):

    if "[" in string and "]" in string:
        e = [
            _evaluate_expression(x.strip())
            for x in string.replace("[", "").replace("]", "").split(",")
            if x.strip() != ""
        ]

        return e

    elif string.startswith("(") and string.endswith(")"):
        e = tuple(
            _evaluate_expression(x.strip())
            for x in string.replace("(", "").replace(")", "").split(",")
            if x.strip() != ""
        )
        return e

    elif string.lower() == "true":
        return True

    elif string.lower() == "false":
        return False

    else:
        return string


def getmembers(module):
    for n in module._columns:
        if n not in ("AllNames", "Next", "First"):
            yield n, getattr(module, n)


def capacitors_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Capacitors)


def fuses_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Fuses)


def generators_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Generators)


def isource_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Isource)


def lines_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Lines)


def loadshape_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.LoadShape)


def loads_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Loads)


def meters_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Meters)


def monitor_to_dataframe(dss=None):
    """
    Return the data from current active monitor as a Pandas DataFrame
    """
    if dss is None:
        from opendssdirect import dss

    if dss.Solution.Mode() in (enums.SolveModes.Harmonic, enums.SolveModes.HarmonicT):
        columns = ['frequency', 'harmonic']
    else:
        columns = ['hour', 'second']

    columns.extend(col.strip() for col in dss.Monitors.Header())
    data = dss.Monitors.AsMatrix()

    if is_pandas_installed:
        return pd.DataFrame(data, columns=columns)
    else:
        warnings.warn(
            "Pandas is not installed. Please see documentation for how to install extra dependencies."
        )
        return dict(zip(columns, data.T))


def monitors_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Monitors)


def pvsystems_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.PVsystems)


def regcontrols_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.RegControls)


def reclosers_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Reclosers)


def relays_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Relays)


def sensors_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Sensors)


def transformers_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Transformers)


def vsources_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.Vsources)


def xycurves_to_dataframe(dss=None):
    if dss is None:
        from opendssdirect import dss
    return to_dataframe(dss.XYCurves)



from typing import Optional, List
from ._utils import api_util
from .Bases import Base
from dss import DSSException, OCPDevType as OCPDevTypeEnum

class ICktElement(Base):
    """
    The (Active)CktElement interface allows accessing some common properties and 
    methods shared across circuit elements in the DSS engine.

    Users can enable specific elements by name or use the dedicated interface 
    (e.g. use `Loads.Name`, `Transformers.First/Next`) and access the properties here.

    If you are new to OpenDSS/AltDSS and this classic interface, please read the following document
    for an overview of the "active element" paradigm used by COM and the classic APIs:
        
    https://dss-extensions.org/classic_api.html#the-active-paradigm
    """

    __slots__ = []

    __name__ = "CktElement"
    _api_prefix = "CktElement"
    _columns = [
        "Name",
        "DisplayName",
        "Handle",
        "GUID",
        "Enabled",
        "NumTerminals",
        "NumPhases",
        "NumConductors",
        "NumControls",
        "NumProperties",
        "AllPropertyNames",
        "AllVariableValues",
        "AllVariableNames",
        "BusNames",
        "NormalAmps",
        "EmergAmps",
        "HasVoltControl",
        "HasSwitchControl",
        "HasOCPDevice",
        "OCPDevType",
        "OCPDevIndex",
        "IsIsolated",
        "EnergyMeter",
        "TotalPowers",
        "YPrim",
        "NodeOrder",
        "Voltages",
        "VoltagesMagAng",
        "SeqVoltages",
        "CplxSeqVoltages",
        "Powers",
        "SeqPowers",
        "Currents",
        "CurrentsMagAng",
        "SeqCurrents",
        "CplxSeqCurrents",
        "Residuals",
        "Losses",
        "PhaseLosses",
        "AllLosses",
    ]

    def Close(self, Term: int, Phs: int = 0) -> None:
        """
        Close the specified terminal and phase, if non-zero, or all conductors at the terminal.

        Original COM help: https://opendss.epri.com/Close1.html
        """
        self._lib.CktElement_Close(Term, Phs)

    def Controller(self, idx: int) -> str:
        """Full name of the i-th controller attached to this element. Ex: str = Controller(2).  See NumControls to determine valid index range"""
        return self._lib.CktElement_Get_Controller(idx)

    def Variable(self, MyVarName: str) -> float:
        """
        If the active element is a PCElement, get the value of a variable by name.
        Otherwise, an exception is raised.

        Original COM help: https://opendss.epri.com/Variable.html
        """
        if not isinstance(MyVarName, bytes):
            MyVarName = MyVarName.encode(self._api_util.codec)
        Code = self._api_util.ffi.new("int32_t*")
        result = self._lib.CktElement_Get_Variable(MyVarName, Code)
        if Code[0] != 0:
            raise DSSException(Code[0], "No variable by this name or not a PCelement.")
        
        return result

    def Variablei(self, Idx: int) -> float:
        """
        If the active element is a PCElement, get the value of a variable by its integer index.
        Otherwise, an exception is raised.

        Original COM help: https://opendss.epri.com/Variablei.html
        """
        Code = self._api_util.ffi.new("int32_t*")
        result = self._lib.CktElement_Get_Variablei(Idx, Code)
        if Code[0] != 0:
            raise DSSException(Code[0], "No variable by this index or not a PCelement.")

        return result

    def setVariableByIndex(self, Idx: int, Value: float) -> None:
        Code = self._api_util.ffi.new("int32_t*")
        self._lib.CktElement_Set_Variablei(Idx, Code, Value)
        return Code[0]

    def setVariableByName(self, Idx: int, Value: float) -> None:
        Code = self._api_util.ffi.new("int32_t*")
        self._lib.CktElement_Set_Variable(Idx, Code, Value)
        return Code[0]

    def IsOpen(self, Term: int, Phs: int = 0) -> bool:
        """
        Indicates if the specified terminal and, optionally, a specific phase conductor is open.

        Provide zero in the `Phs` argument to check if any conductor of the terminal `Term` is open.

        Provide a non-zero phase number in `Phs` to check if a specific phase conductor is open.

        Original COM help: https://opendss.epri.com/IsOpen.html
        """
        return self._lib.CktElement_IsOpen(Term, Phs)

    def Open(self, Term: int, Phs: int = 0) -> None:
        """
        Open the specified terminal and phase, if non-zero, or all conductors at the terminal.

        Original COM help: https://opendss.epri.com/Open1.html
        """
        self._lib.CktElement_Open(Term, Phs)

    def AllPropertyNames(self) -> List[str]:
        """
        Array containing all property names of the active device.

        Original COM help: https://opendss.epri.com/AllPropertyNames.html
        """
        return self._lib.CktElement_Get_AllPropertyNames()

    def AllVariableNames(self) -> List[str]:
        """
        Array of strings listing all the published state variable names.
        Valid only for PCElements.

        Original COM help: https://opendss.epri.com/AllVariableNames.html
        """
        return self._lib.CktElement_Get_AllVariableNames()

    def AllVariableValues(self):
        """
        Array of doubles. Values of state variables of active element if PC element.
        Valid only for PCElements.

        Original COM help: https://opendss.epri.com/AllVariableValues.html
        """
        return self._lib.CktElement_Get_AllVariableValues_GR()

    def BusNames(self, *args, removeNodes: bool = False) -> List[str]:
        """
        Bus definitions to which each terminal is connected.

        In getter mode, the keyword argument `removeNodes` is an **API Extension**.
        Use it to get only the bus names, without the connection/node specification, if present.

        Original COM help: https://opendss.epri.com/BusNames.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CktElement_Get_BusNames(removeNodes)

        # Setter
        (Value,) = args
        self._set_string_array(self._lib.CktElement_Set_BusNames, Value)

    def CplxSeqCurrents(self):
        """
        Complex double array of Sequence Currents for all conductors of all terminals of active circuit element.

        Original COM help: https://opendss.epri.com/CplxSeqCurrents.html
        """
        return self._lib.CktElement_Get_CplxSeqCurrents_GR()

    def CplxSeqVoltages(self):
        """
        Complex double array of Sequence Voltage for all terminals of active circuit element.

        Original COM help: https://opendss.epri.com/CplxSeqVoltages1.html
        """
        return self._lib.CktElement_Get_CplxSeqVoltages_GR()

    def Currents(self):
        """
        Complex array of currents into each conductor of each terminal

        Original COM help: https://opendss.epri.com/Currents1.html
        """
        return self._lib.CktElement_Get_Currents_GR()

    def CurrentsMagAng(self):
        """
        Currents in magnitude, angle (degrees) format as an array of doubles.

        Original COM help: https://opendss.epri.com/CurrentsMagAng.html
        """
        return self._lib.CktElement_Get_CurrentsMagAng_GR()

    def DisplayName(self, *args):
        """
        Display name of the object (not necessarily unique)

        Original COM help: https://opendss.epri.com/DisplayName.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CktElement_Get_DisplayName()

        # Setter
        (Value,) = args
        self._lib.CktElement_Set_DisplayName(Value)

    def EmergAmps(self, *args):
        """
        Emergency Ampere Rating for PD elements

        Original COM help: https://opendss.epri.com/EmergAmps.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CktElement_Get_EmergAmps()

        # Setter
        (Value,) = args
        self._lib.CktElement_Set_EmergAmps(Value)

    def Enabled(self, *args) -> Optional[bool]:
        """
        Boolean indicating that element is currently in the circuit.

        Original COM help: https://opendss.epri.com/Enabled.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CktElement_Get_Enabled()

        # Setter
        (Value,) = args
        self._lib.CktElement_Set_Enabled(Value)

    def EnergyMeter(self) -> str:
        """
        Name of the Energy Meter this element is assigned to.

        *Requires an energy meter with an updated zone.*

        Original COM help: https://opendss.epri.com/EnergyMeter.html
        """
        return self._lib.CktElement_Get_EnergyMeter()

    def GUID(self) -> str:
        """
        GUID/UUID for this object.

        Original COM help: https://opendss.epri.com/GUID.html
        """
        return self._lib.CktElement_Get_GUID()

    def Handle(self) -> int:
        """
        Index of this element into the circuit's element list.

        Original COM help: https://opendss.epri.com/Handle.html
        """
        return self._lib.CktElement_Get_Handle()

    def HasOCPDevice(self) -> bool:
        """
        True if a recloser, relay, or fuse controlling this ckt element. OCP = Overcurrent Protection

        Original COM help: https://opendss.epri.com/HasOCPDevice.html
        """
        return self._lib.CktElement_Get_HasOCPDevice()

    def HasSwitchControl(self) -> bool:
        """
        True if this element has a SwtControl attached.

        Original COM help: https://opendss.epri.com/HasSwitchControl.html
        """
        return self._lib.CktElement_Get_HasSwitchControl()

    def HasVoltControl(self) -> bool:
        """
        True if this element has a CapControl or RegControl attached.

        Original COM help: https://opendss.epri.com/HasVoltControl.html
        """
        return self._lib.CktElement_Get_HasVoltControl()

    def Losses(self):
        """
        Total losses in the element: two-element double array (complex), in VA (watts, vars)

        Original COM help: https://opendss.epri.com/Losses1.html
        """
        return self._lib.CktElement_Get_Losses_GR()

    def Name(self) -> str:
        """
        Full Name of Active Circuit Element

        Original COM help: https://opendss.epri.com/Name4.html
        """
        return self._lib.CktElement_Get_Name()

    def NodeOrder(self):
        """
        Array of integer containing the node numbers (representing phases, for example) for each conductor of each terminal.

        Be sure to run a solution to initialize the values after the circuit is created or modified.

        Original COM help: https://opendss.epri.com/NodeOrder.html
        """
        return self._lib.CktElement_Get_NodeOrder_GR()

    def NormalAmps(self, *args):
        """
        Normal ampere rating for PD Elements

        Original COM help: https://opendss.epri.com/NormalAmps.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CktElement_Get_NormalAmps()

        # Setter
        (Value,) = args
        self._lib.CktElement_Set_NormalAmps(Value)

    def NumConductors(self) -> int:
        """
        Number of Conductors per Terminal

        Original COM help: https://opendss.epri.com/NumConductors.html
        """
        return self._lib.CktElement_Get_NumConductors()

    def NumControls(self) -> int:
        """
        Number of controls connected to this device.
        Use to determine valid range for index into Controller array.

        Original COM help: https://opendss.epri.com/NumControls.html
        """
        return self._lib.CktElement_Get_NumControls()

    def NumPhases(self) -> int:
        """
        Number of Phases

        Original COM help: https://opendss.epri.com/NumPhases.html
        """
        return self._lib.CktElement_Get_NumPhases()

    def NumProperties(self) -> int:
        """
        Number of Properties this Circuit Element.

        Original COM help: https://opendss.epri.com/NumProperties.html
        """
        return self._lib.CktElement_Get_NumProperties()

    def NumTerminals(self) -> int:
        """
        Number of terminals in this Circuit Element

        Original COM help: https://opendss.epri.com/NumTerminals.html
        """
        return self._lib.CktElement_Get_NumTerminals()

    def OCPDevIndex(self) -> int:
        """
        Index into Controller list of OCP Device controlling this CktElement

        Original COM help: https://opendss.epri.com/OCPDevIndex.html
        """
        return self._lib.CktElement_Get_OCPDevIndex()

    def OCPDevType(self) -> int:
        """
        0=None; 1=Fuse; 2=Recloser; 3=Relay;  Type of OCP controller device

        Original COM help: https://opendss.epri.com/OCPDevType.html
        """
        return OCPDevTypeEnum(self._lib.CktElement_Get_OCPDevType())

    def PhaseLosses(self):
        """
        Complex array of losses (kVA) by phase

        Original COM help: https://opendss.epri.com/PhaseLosses.html
        """
        return self._lib.CktElement_Get_PhaseLosses_GR()

    def AllLosses(self):
        """
        Complex array with the losses by type (total losses, load losses, no-load losses), in VA, for the active circuit element.

        Added in May 2025. Same as `LossesByType` introduced for Transformers in AltDSS/DSS C-API in May 2019.
        """
        return self._lib.CktElement_Get_AllLosses_GR()

    def Powers(self):
        """
        Complex array of powers (kVA) into each conductor of each terminal

        Original COM help: https://opendss.epri.com/Powers.html
        """
        return self._lib.CktElement_Get_Powers_GR()

    def Residuals(self):
        """
        Residual currents for each terminal: (magnitude, angle in degrees)

        Original COM help: https://opendss.epri.com/Residuals.html
        """
        return self._lib.CktElement_Get_Residuals_GR()

    def SeqCurrents(self):
        """
        Double array of symmetrical component currents (magnitudes only) into each 3-phase terminal

        Original COM help: https://opendss.epri.com/SeqCurrents.html
        """
        return self._lib.CktElement_Get_SeqCurrents_GR()

    def SeqPowers(self):
        """
        Complex array of sequence powers (kW, kvar) into each 3-phase terminal

        Original COM help: https://opendss.epri.com/SeqPowers.html
        """
        return self._lib.CktElement_Get_SeqPowers_GR()

    def SeqVoltages(self):
        """
        Double array of symmetrical component voltages (magnitudes only) at each 3-phase terminal

        Original COM help: https://opendss.epri.com/SeqVoltages1.html
        """
        return self._lib.CktElement_Get_SeqVoltages_GR()

    def Voltages(self):
        """
        Complex array of voltages at terminals

        Original COM help: https://opendss.epri.com/Voltages1.html
        """
        return self._lib.CktElement_Get_Voltages_GR()

    def VoltagesMagAng(self):
        """
        Voltages at each conductor in magnitude, angle form as array of doubles.

        Original COM help: https://opendss.epri.com/VoltagesMagAng.html
        """
        return self._lib.CktElement_Get_VoltagesMagAng_GR()

    def YPrim(self):
        """
        YPrim matrix, column order, complex numbers

        Original COM help: https://opendss.epri.com/Yprim.html
        """
        return self._lib.CktElement_Get_Yprim_GR()

    def YprimOrder(self) -> int:
        """
        Order (size) of the active circuit element's primite Y matrix (Yprim), typically `NumConductors * NumTerminals`

        **(API Extension)**
        """
        return self._lib.CktElement_Get_YprimOrder()

    def IsIsolated(self) -> bool:
        """
        Returns true if the current active element is isolated.
        Note that this only fetches the current value. See also the Topology interface.

        **(API Extension)**
        """
        return self._lib.CktElement_Get_IsIsolated()

    def TotalPowers(self):
        """
        Returns an array with the total powers (complex, kVA) at ALL terminals of the active circuit element.

        Original COM help: https://opendss.epri.com/TotalPowers.html
        """
        return self._lib.CktElement_Get_TotalPowers_GR()

    def NodeRef(self):
        """
        Array of integers, a copy of the internal NodeRef of the CktElement.

        Be sure to run a solution to initialize the values after the circuit is created or modified.

        **(API Extension)**
        """
        return self._lib.CktElement_Get_NodeRef_GR()


_CktElement = ICktElement(api_util)

# For backwards compatibility, bind to the default instance
AllLosses = _CktElement.AllLosses
AllPropertyNames = _CktElement.AllPropertyNames
AllVariableNames = _CktElement.AllVariableNames
AllVariableValues = _CktElement.AllVariableValues
BusNames = _CktElement.BusNames
Close = _CktElement.Close
Controller = _CktElement.Controller
CplxSeqCurrents = _CktElement.CplxSeqCurrents
CplxSeqVoltages = _CktElement.CplxSeqVoltages
Currents = _CktElement.Currents
CurrentsMagAng = _CktElement.CurrentsMagAng
DisplayName = _CktElement.DisplayName
EmergAmps = _CktElement.EmergAmps
Enabled = _CktElement.Enabled
EnergyMeter = _CktElement.EnergyMeter
GUID = _CktElement.GUID
Handle = _CktElement.Handle
HasOCPDevice = _CktElement.HasOCPDevice
HasSwitchControl = _CktElement.HasSwitchControl
HasVoltControl = _CktElement.HasVoltControl
IsIsolated = _CktElement.IsIsolated
IsOpen = _CktElement.IsOpen
Losses = _CktElement.Losses
Name = _CktElement.Name
NodeOrder = _CktElement.NodeOrder
NodeRef = _CktElement.NodeRef
NormalAmps = _CktElement.NormalAmps
NumConductors = _CktElement.NumConductors
NumControls = _CktElement.NumControls
NumPhases = _CktElement.NumPhases
NumProperties = _CktElement.NumProperties
NumTerminals = _CktElement.NumTerminals
OCPDevIndex = _CktElement.OCPDevIndex
OCPDevType = _CktElement.OCPDevType
Open = _CktElement.Open
PhaseLosses = _CktElement.PhaseLosses
Powers = _CktElement.Powers
Residuals = _CktElement.Residuals
SeqCurrents = _CktElement.SeqCurrents
SeqPowers = _CktElement.SeqPowers
SeqVoltages = _CktElement.SeqVoltages
TotalPowers = _CktElement.TotalPowers
Variable = _CktElement.Variable
Variablei = _CktElement.Variablei
Voltages = _CktElement.Voltages
VoltagesMagAng = _CktElement.VoltagesMagAng
YPrim = _CktElement.YPrim
YprimOrder = _CktElement.YprimOrder
setVariableByIndex = _CktElement.setVariableByIndex
setVariableByName = _CktElement.setVariableByName
_columns = _CktElement._columns
__all__ = [
    "AllLosses",
    "AllPropertyNames",
    "AllVariableNames",
    "AllVariableValues",
    "BusNames",
    "Close",
    "Controller",
    "CplxSeqCurrents",
    "CplxSeqVoltages",
    "Currents",
    "CurrentsMagAng",
    "DisplayName",
    "EmergAmps",
    "Enabled",
    "EnergyMeter",
    "GUID",
    "Handle",
    "HasOCPDevice",
    "HasSwitchControl",
    "HasVoltControl",
    "IsIsolated",
    "IsOpen",
    "Losses",
    "Name",
    "NodeOrder",
    "NodeRef",
    "NormalAmps",
    "NumConductors",
    "NumControls",
    "NumPhases",
    "NumProperties",
    "NumTerminals",
    "OCPDevIndex",
    "OCPDevType",
    "Open",
    "PhaseLosses",
    "Powers",
    "Residuals",
    "SeqCurrents",
    "SeqPowers",
    "SeqVoltages",
    "TotalPowers",
    "Variable",
    "Variablei",
    "Voltages",
    "VoltagesMagAng",
    "YPrim",
    "YprimOrder",
    "setVariableByIndex",
    "setVariableByName",
]



from ._utils import api_util
from .Bases import Base
from dss.enums import DSSPropertyNameStyle, CktModels

class SettingsContext:
    def __init__(self, settings):
        self._settings = settings
    
    def __enter__(self):
        # Using try...except since the official engine doesn't implement these.
        # Only a few are (and can be) implemented through Oddie.
        try:
            self._AdvancedTypes = self._settings.AdvancedTypes()
        except:
            pass

        try:
            self._CompatFlags = self._settings.CompatFlags()
        except:
            pass

        try:
            self._IterateDisabled = self._settings.IterateDisabled()
        except:
            pass

        try:
            self._PreferLists = self._settings.PreferLists()
        except:
            pass

        try:
            self._SkipCommands = self._settings.SkipCommands()
        except:
            pass

        try:
            self._SkipFileRegExp = self._settings.SkipFileRegExp()
        except:
            pass

        return self._settings
        
    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self._settings.AdvancedTypes(self._AdvancedTypes)
        except:
            pass

        try:
            self._settings.CompatFlags(self._CompatFlags)
        except:
            pass

        try:
            self._settings.IterateDisabled(self._IterateDisabled)
        except:
            pass

        try:
            self._settings.PreferLists(self._PreferLists)
        except:
            pass

        try:
            self._settings.SkipCommands(self._SkipCommands)
        except:
            pass

        try:
            self._settings.SkipFileRegExp(self._SkipFileRegExp)
        except:
            pass


class ISettings(Base):
    __slots__ = [
        '_command_dict',
    ]

    __name__ = "Settings"
    _api_prefix = "Settings"
    _columns = [
        "Trapezoidal",
        "LossRegs",
        "VoltageBases",
        "ZoneLock",
        "EmergVminpu",
        "PriceSignal",
        "CktModel",
        "UERegs",
        "UEWeight",
        "PriceCurve",
        "NormVminpu",
        "LossWeight",
        "EmergVmaxpu",
        "AutoBusList",
        "NormVmaxpu",
        "AllowDuplicates",
        "ControlTrace",
        "LoadsTerminalCheck",
        "IterateDisabled",
    ]

    def __init__(self, api_util):
        Base.__init__(self, api_util)
        num_commands = self._lib.DSS_Executive_Get_NumCommands()
        object.__setattr__(self, '_frozen_attrs', False)
        self._command_dict = {
            self._lib.DSS_Executive_Get_Command(i).lower(): i
            for i in range(1, num_commands + 1)
        }
        object.__setattr__(self, '_frozen_attrs', True)


    def Context(self) -> SettingsContext:
        """
        Returns a Settings context manager. 
        The context manager saves the values of the tracker settings on enter, 
        restoring them on exit. This allows code to change the settings within 
        the context block and they are restored to the initial values automatically.

        Note: this context manager target OpenDSSDirect.py settings. Use the equivalent for DSS-Python.
        A few settings are shared at engine level.

        Settings tracked:
        - AdvancedTypes
        - CompatFlags
        - IterateDisabled
        - PreferLists
        - SkipCommands
        - SkipFileRegExp
        """
        return SettingsContext(self)

    def AllowDuplicates(self, *args):
        """
        Designates whether to allow duplicate names of objects

        False by default.

        **NOTE**: for DSS-Extensions, we are considering removing this option in a future
        release since it has performance impacts even when not used.
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_AllowDuplicates()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_AllowDuplicates(Value)

    def AutoBusList(self, *args):
        """
        List of Buses or (File=xxxx) syntax for the AutoAdd solution mode.

        Original COM help: https://opendss.epri.com/AutoBusList.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_AutoBusList()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_AutoBusList(Value)

    def CktModel(self, *args):
        """
        Indicate if the circuit model is positive sequence.

        Original COM help: https://opendss.epri.com/CktModel.html
        """
        # Getter
        if len(args) == 0:
            return CktModels(self._lib.Settings_Get_CktModel())

        # Setter
        (Value,) = args
        self._lib.Settings_Set_CktModel(Value)

    def ControlTrace(self, *args):
        """
        Denotes whether to trace the control actions to a file.

        Original COM help: https://opendss.epri.com/ControlTrace.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_ControlTrace()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_ControlTrace(Value)

    def EmergVmaxpu(self, *args):
        """
        Per Unit maximum voltage for Emergency conditions.

        Original COM help: https://opendss.epri.com/EmergVmaxpu.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_EmergVmaxpu()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_EmergVmaxpu(Value)

    def EmergVminpu(self, *args):
        """
        Per Unit minimum voltage for Emergency conditions.

        Original COM help: https://opendss.epri.com/EmergVminpu.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_EmergVminpu()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_EmergVminpu(Value)

    def LossRegs(self, *args):
        """
        Integer array defining which energy meter registers to use for computing losses

        Original COM help: https://opendss.epri.com/LossRegs.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_LossRegs_GR()

        # Setter
        (Value,) = args
        Value, ValuePtr, ValueCount = self._prepare_int32_array(Value)
        self._lib.Settings_Set_LossRegs(ValuePtr, ValueCount)

    def LossWeight(self, *args):
        """
        Weighting factor applied to Loss register values.

        Original COM help: https://opendss.epri.com/LossWeight.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_LossWeight()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_LossWeight(Value)

    def NormVmaxpu(self, *args):
        """
        Per Unit maximum voltage for Normal conditions.

        Original COM help: https://opendss.epri.com/NormVmaxpu.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_NormVmaxpu()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_NormVmaxpu(Value)

    def NormVminpu(self, *args):
        """
        Per Unit minimum voltage for Normal conditions.

        Original COM help: https://opendss.epri.com/NormVminpu.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_NormVminpu()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_NormVminpu(Value)

    def PriceCurve(self, *args):
        """
        Name of LoadShape object that serves as the source of price signal data for yearly simulations, etc.

        Original COM help: https://opendss.epri.com/PriceCurve.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_PriceCurve()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_PriceCurve(Value)

    def PriceSignal(self, *args):
        """
        Price Signal for the Circuit

        Original COM help: https://opendss.epri.com/PriceSignal.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_PriceSignal()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_PriceSignal(Value)

    def Trapezoidal(self, *args):
        """
        Gets value of trapezoidal integration flag in energy meters. Defaults to `False`.

        Original COM help: https://opendss.epri.com/Trapezoidal.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_Trapezoidal()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_Trapezoidal(Value)

    def UERegs(self, *args):
        """
        Array of Integers defining energy meter registers to use for computing UE

        Original COM help: https://opendss.epri.com/UEregs.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_UEregs_GR()

        # Setter
        (Value,) = args
        Value, ValuePtr, ValueCount = self._prepare_int32_array(Value)
        self._lib.Settings_Set_UEregs(ValuePtr, ValueCount)

    def UEWeight(self, *args):
        """
        Weighting factor applied to UE register values.

        Original COM help: https://opendss.epri.com/UEweight.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_UEweight()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_UEweight(Value)

    def VoltageBases(self, *args):
        """
        Array of doubles defining the legal voltage bases in kV L-L

        Original COM help: https://opendss.epri.com/VoltageBases.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_VoltageBases_GR()

        # Setter
        (Value,) = args
        Value, ValuePtr, ValueCount = self._prepare_float64_array(Value)
        self._lib.Settings_Set_VoltageBases(ValuePtr, ValueCount)

    def ZoneLock(self, *args):
        """
        Locks Zones on energy meters to prevent rebuilding if a circuit change occurs.

        Original COM help: https://opendss.epri.com/ZoneLock.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_ZoneLock()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_ZoneLock(Value)

    def AllocationFactors(self, Value):
        """(write-only) Sets all load allocation factors for all loads defined by XFKVA property to this value."""
        self._lib.Settings_Set_AllocationFactors(Value)

    def LoadsTerminalCheck(self, *args):
        """
        Controls whether the terminals are checked when updating the currents in Load component. Defaults to True.
        If the loads are guaranteed to have their terminals closed throughout the simulation, this can be set to False to save some time.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_LoadsTerminalCheck()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_LoadsTerminalCheck(Value)

    def IterateDisabled(self, *args):
        """
        Controls whether `First`/`Next` iteration includes or skips disabled circuit elements.
        The default behavior from OpenDSS is to skip those. The user can still activate the element by name or index.

        The default value for IterateDisabled is 0, keeping the original behavior.
        Set it to 1 (or `True`) to include disabled elements.
        Other numeric values are reserved for other potential behaviors.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_IterateDisabled()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_IterateDisabled(Value)

    def SetPropertyNameStyle(self, value):
        """
        Switch the property names according to the target style.

        Use this method for compatibility with code that doesn't consider that
        OpenDSS is case insensitive. Check the enumeration for more:
        [DSSPropertyNameStyle](#dss_python_backend.enums.DSSPropertyNameStyle)

        **(API Extension)**
        """
        self._lib.Settings_SetPropertyNameStyle(value)


    def SkipFileRegExp(self, *args):
        """
        Regular expression pattern to skip files.

        If a file name as provided in the input for the `Redirect` and `Compile` commands
        matches the regular expression pattern, it is skipped (the file is not read nor
        commands contained in the file are executed).

        Set to an empty string to reset/disable the filter.

        Case-insensitive.
        See https://regex.sorokin.engineer/en/latest/regular_expressions.html for information on
        the expression syntax and options.

        Even if the `clear` command is included in `Settings.SkipCommands()`, the `DSS.ClearAll()` method can 
        still be called. It resets both skip settings, `SkipCommands()` and `SkipFileRegExp()`.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_SkipFileRegExp()

        # Setter
        (Value,) = args
        self._lib.Settings_Set_SkipFileRegExp(Value or "")


    def SkipCommands(self, *args):
        """
        List of commands to skip

        List of strings representing the command names to skip when processing DSS text commands or files.

        If the `clear` command is included in `Settings.SkipCommands()`, the `ClearAll()` method can
        still be called and it will reset both skip settings, `SkipCommands` and `SkipFileRegExp`.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return [
                self._lib.DSS_Executive_Get_Command(i)
                for i in self._lib.Settings_Get_SkipCommands_GR()
            ]

        # Setter
        (Value,) = args
        if len(Value) != 0 and isinstance(Value[0], str):
            # map command names to integer codes
            Value = [self._command_dict[cmd_name] for cmd_name in Value]

        Value, ValuePtr, ValueCount = self._prepare_int32_array(Value)
        self._lib.Settings_Set_SkipCommands(ValuePtr, ValueCount)


    def AdvancedTypes(self, *args):
        """
        When enabled, there are **two side-effects**:

        - **Per DSS Context:** Complex arrays and complex numbers can be returned and consumed by the Python API.
        - **Global effect:** The low-level API provides matrix dimensions when available (`EnableArrayDimensions` is enabled).

        As a result, for example, `OpenDSSDirect.CktElement.Yprim()` is returned as a complex matrix instead
        of a plain array.

        When disabled, the legacy plain arrays are used and complex numbers cannot be consumed by the Python API.

        *Defaults to **False** for backwards compatibility.*

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.advanced_types

        # Setter
        (Value,) = args
        self._lib.advanced_types = bool(Value)


    def CompatFlags(self, *args):
        """
        Controls some compatibility flags introduced to toggle some behavior from EPRI's OpenDSS.

        **THE FLAGS ARE GLOBAL, affecting all AltDSS engines in the process.**
        CompatFlags for Oddie-loaded instances (OpenDSS and OpenDSS-C engines) are handled by the Oddie code itself,
        so it is global for each Oddie library.

        These flags may change for each version of DSS C-API, but the same value will not be reused. That is,
        when we remove a compatibility flag, it will have no effect but will also not affect anything else
        besides raising an error if the user tries to toggle a flag that was available in a previous version.

        We expect to keep a very limited number of flags. Since the flags are more transient than the other
        options/flags, it was preferred to add this generic function instead of a separate function per
        flag.

        See the enumeration `DSSCompatFlags` for available flags, including description.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_CompatFlags()

        # Setter
        (Value,) = args
        self._lib.DSS_Set_CompatFlags(Value)

    def PreferLists(self, *args):
        """
        Enable this setting to use lists instead of NumPy arrays, where it makes sense.

        This was added for better for compatibility with the COM packages (`comtypes` and  `win32com` use lists
        and tuples by default) and the original OpenDSSDirect.py releases.

        Current releases of OpenDSSDirect.py, DSS-Python, and AltDSS(-Python) all use NumPy arrays by default.
        Users can also activate the related `AdvancedTypes` for a richer experience.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.prefer_lists

        # Setter
        (value,) = args
        self._lib.prefer_lists = value

    def COMErrorResults(self, *args):
        """
        If enabled, in case of errors or empty arrays, the API returns arrays with values compatible with
        EPRI's OpenDSS COM interface.

        For example, consider the method/property `Loads.ZIPV()`. If there is no active circuit or active load element:

        - In the disabled state (`COMErrorResults(False)`), the function will return "[]", an array with 0 elements.
        - In the enabled state (`COMErrorResults(True)`), the function will return "[0.0]" instead. This should
        be compatible with the return value of EPRI's OpenDSS COM interface.

        Defaults to false (disabled state) in AltDSS since the v0.15.x series.

        This does not affect the results when using EPRI's OpenDSS distribution through Oddie.

        This can also be set through the environment variable `DSS_CAPI_COM_DEFAULTS`. Setting it to 1 enables
        the legacy/COM behavior. The value can be toggled through the API at any time.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_COMErrorResults()

        # Setter
        (Value,) = args
        self._lib.DSS_Set_COMErrorResults(Value)


    def AllowDOScmd(self, *args):
        """
        If enabled, the `DOScmd` command is allowed. Otherwise, an error is reported if the user tries to use it.

        Defaults to False/0 (disabled state). Users should consider DOScmd deprecated on DSS-Extensions.

        This can also be set through the environment variable `DSS_CAPI_ALLOW_DOSCMD`. Setting it to 1 enables
        the command.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_AllowDOScmd()

        # Setter
        (Value,) = args
        self._lib.DSS_Set_AllowDOScmd(Value)

    def AllowChangeDir(self, *args):
        """
        If disabled, the engine will not change the active working directory during execution. E.g. a "compile"
        command will not "chdir" to the file path.

        If you have issues with long paths, enabling this might help in some scenarios.

        Defaults to True (allow changes, backwards compatible) in the 0.10.x versions of DSS C-API.
        This might change to False in future versions.

        This can also be set through the environment variable `DSS_CAPI_ALLOW_CHANGE_DIR`. Set it to 0 to
        disallow changing the active working directory.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_AllowChangeDir()

        # Setter
        (Value,) = args
        self._lib.DSS_Set_AllowChangeDir(Value)

    def AllowEditor(self, *args):
        """
        Gets/sets whether running the external editor for "Show" is allowed

        AllowEditor controls whether the external editor is used in commands like "Show".
        If you set to 0 (false), the editor is not executed. Note that other side effects,
        such as the creation of files, are not affected.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_AllowEditor()

        # Setter
        (value,) = args
        self._lib.DSS_Set_AllowEditor(value)

    def PreserveCase(self, *args):
        """
        Gets/sets whether the engine tries to preserve original names

        When enabled, bus and element names in many of the API functions, reports and
        exports are kept as provided by the user, without applying lower or upper case
        transformations.

        Note that, even when enabled, the engine is still case-insensitive.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Settings_Get_Flag(1)

        # Setter
        (value,) = args
        self._lib.Settings_Set_Flag(1, value)


_Settings = ISettings(api_util)

# For backwards compatibility, bind to the default instance
AdvancedTypes = _Settings.AdvancedTypes
AllocationFactors = _Settings.AllocationFactors
AllowChangeDir = _Settings.AllowChangeDir
AllowDOScmd = _Settings.AllowDOScmd
AllowDuplicates = _Settings.AllowDuplicates
AllowEditor = _Settings.AllowEditor
AutoBusList = _Settings.AutoBusList
COMErrorResults = _Settings.COMErrorResults
CktModel = _Settings.CktModel
CompatFlags = _Settings.CompatFlags
Context = _Settings.Context
ControlTrace = _Settings.ControlTrace
EmergVmaxpu = _Settings.EmergVmaxpu
EmergVminpu = _Settings.EmergVminpu
IterateDisabled = _Settings.IterateDisabled
LoadsTerminalCheck = _Settings.LoadsTerminalCheck
LossRegs = _Settings.LossRegs
LossWeight = _Settings.LossWeight
NormVmaxpu = _Settings.NormVmaxpu
NormVminpu = _Settings.NormVminpu
PreferLists = _Settings.PreferLists
PreserveCase = _Settings.PreserveCase
PriceCurve = _Settings.PriceCurve
PriceSignal = _Settings.PriceSignal
SetPropertyNameStyle = _Settings.SetPropertyNameStyle
SkipCommands = _Settings.SkipCommands
SkipFileRegExp = _Settings.SkipFileRegExp
Trapezoidal = _Settings.Trapezoidal
UERegs = _Settings.UERegs
UEWeight = _Settings.UEWeight
VoltageBases = _Settings.VoltageBases
ZoneLock = _Settings.ZoneLock
_columns = _Settings._columns
__all__ = [
    "AdvancedTypes",
    "AllocationFactors",
    "AllowChangeDir",
    "AllowDOScmd",
    "AllowDuplicates",
    "AllowEditor",
    "AutoBusList",
    "COMErrorResults",
    "CktModel",
    "CompatFlags",
    "Context",
    "ControlTrace",
    "EmergVmaxpu",
    "EmergVminpu",
    "IterateDisabled",
    "LoadsTerminalCheck",
    "LossRegs",
    "LossWeight",
    "NormVmaxpu",
    "NormVminpu",
    "PreferLists",
    "PreserveCase",
    "PriceCurve",
    "PriceSignal",
    "SetPropertyNameStyle",
    "SkipCommands",
    "SkipFileRegExp",
    "Trapezoidal",
    "UERegs",
    "UEWeight",
    "VoltageBases",
    "ZoneLock",
]



from ._utils import api_util
from .Bases import Iterable
from dss import GeneratorStatus


class IGenerators(Iterable):
    __slots__ = []

    __name__ = "Generators"
    _api_prefix = "Generators"
    _columns = [
        "Name",
        "Idx",
        "ForcedON",
        "Model",
        "Phases",
        "PF",
        "kVARated",
        "kV",
        "kW",
        "kvar",
        "Vminpu",
        "Vmaxpu",
        "RegisterNames",
        "RegisterValues",
        "Bus1",
        "Class",
        "kva",
        "IsDelta",
        "Status",
        "daily",
        "duty",
        "Yearly",
    ]

    def ForcedON(self, *args):
        """
        Indicates whether the generator is forced ON regardless of other dispatch criteria.

        Original COM help: https://opendss.epri.com/ForcedON.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_ForcedON()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_ForcedON(Value)

    def Model(self, *args):
        """
        Generator Model

        Original COM help: https://opendss.epri.com/Model.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_Model()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_Model(Value)

    def PF(self, *args):
        """
        Power factor (pos. = producing vars). Updates kvar based on present kW value.

        Original COM help: https://opendss.epri.com/PF.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_PF()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_PF(Value)

    def Phases(self, *args):
        """
        Number of phases

        Original COM help: https://opendss.epri.com/Phases.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_Phases()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_Phases(Value)

    def RegisterNames(self):
        """
        Array of Names of all generator energy meter registers

        See also the enum `GeneratorRegisters`.
        """
        return self._lib.Generators_Get_RegisterNames()

    def RegisterValues(self):
        """
        Array of values in generator energy meter registers.

        Original COM help: https://opendss.epri.com/RegisterValues.html
        """
        return self._lib.Generators_Get_RegisterValues_GR()

    def Vmaxpu(self, *args):
        """
        Vmaxpu for generator model

        Original COM help: https://opendss.epri.com/Vmaxpu.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_Vmaxpu()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_Vmaxpu(Value)

    def Vminpu(self, *args):
        """
        Vminpu for Generator model

        Original COM help: https://opendss.epri.com/Vminpu.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_Vminpu()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_Vminpu(Value)

    def kV(self, *args):
        """
        Voltage base for the active generator, kV

        Original COM help: https://opendss.epri.com/kV1.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_kV()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_kV(Value)

    def kVARated(self, *args):
        """
        kVA rating of the generator

        Original COM help: https://opendss.epri.com/kVArated.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_kVArated()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_kVArated(Value)

    def kW(self, *args):
        """
        kW output for the active generator. kvar is updated for current power factor.

        Original COM help: https://opendss.epri.com/kW.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_kW()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_kW(Value)

    def kvar(self, *args):
        """
        kvar output for the active generator. Updates power factor based on present kW value.

        Original COM help: https://opendss.epri.com/kvar.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_kvar()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_kvar(Value)

    def daily(self, *args):
        """
        Name of the loadshape for a daily generation profile.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_daily()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_daily(Value)

    def duty(self, *args):
        """
        Name of the loadshape for a duty cycle simulation.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_duty()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_duty(Value)

    def Yearly(self, *args):
        """
        Name of yearly loadshape

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_Yearly()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_Yearly(Value)

    def Status(self, *args):
        """
        Response to dispatch multipliers: Fixed=1 (dispatch multipliers do not apply), Variable=0 (follows curves).

        Related enumeration: GeneratorStatus

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return GeneratorStatus(self._lib.Generators_Get_Status())

        # Setter
        (Value,) = args
        self._lib.Generators_Set_Status(Value)

    def IsDelta(self, *args):
        """
        Generator connection. True/1 if delta connection, False/0 if wye.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_IsDelta()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_IsDelta(Value)

    def kva(self, *args):
        """
        kVA rating of electrical machine. Applied to machine or inverter definition for Dynamics mode solutions.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_kva()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_kva(Value)

    def Class(self, *args):
        """
        An arbitrary integer number representing the class of Generator so that Generator values may be segregated by class.
        No effect on the solution.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_Class_()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_Class_(Value)

    def Bus1(self, *args):
        """
        Bus to which the Generator is connected. May include specific node specification.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Generators_Get_Bus1()

        # Setter
        (Value,) = args
        self._lib.Generators_Set_Bus1(Value)


_Generators = IGenerators(api_util)

# For backwards compatibility, bind to the default instance
AllNames = _Generators.AllNames
Bus1 = _Generators.Bus1
Class = _Generators.Class
Count = _Generators.Count
First = _Generators.First
ForcedON = _Generators.ForcedON
Idx = _Generators.Idx
IsDelta = _Generators.IsDelta
Model = _Generators.Model
Name = _Generators.Name
Next = _Generators.Next
PF = _Generators.PF
Phases = _Generators.Phases
RegisterNames = _Generators.RegisterNames
RegisterValues = _Generators.RegisterValues
Status = _Generators.Status
Vmaxpu = _Generators.Vmaxpu
Vminpu = _Generators.Vminpu
Yearly = _Generators.Yearly
daily = _Generators.daily
duty = _Generators.duty
kV = _Generators.kV
kVARated = _Generators.kVARated
kW = _Generators.kW
kva = _Generators.kva
kvar = _Generators.kvar
_columns = _Generators._columns
__all__ = [
    "AllNames",
    "Bus1",
    "Class",
    "Count",
    "First",
    "ForcedON",
    "Idx",
    "IsDelta",
    "Model",
    "Name",
    "Next",
    "PF",
    "Phases",
    "RegisterNames",
    "RegisterValues",
    "Status",
    "Vmaxpu",
    "Vminpu",
    "Yearly",
    "daily",
    "duty",
    "kV",
    "kVARated",
    "kW",
    "kva",
    "kvar",
]



from ._utils import api_util
from .Bases import Iterable
from dss import LineUnits


class IWireData(Iterable):
    """
    WireData objects

    *Not available when using EPRI's OpenDSS distribution.*

    **(API Extension)**
    """

    __name__ = "WireData"
    _api_prefix = "WireData"
    __slots__ = []
    _columns = [
        "Name",
        "Idx",
        "NormAmps",
        "EmergAmps",
        "Rdc",
        "Rac",
        "ResistanceUnits",
        "GMRac",
        "GMRUnits",
        "Radius",
        "Diameter",
        "RadiusUnits",
        "CapRadius",
    ]

    def EmergAmps(self, *args):
        """Emergency ampere rating"""
        # Getter
        if len(args) == 0:
            return self._lib.WireData_Get_EmergAmps()

        # Setter
        (Value,) = args
        self._lib.WireData_Set_EmergAmps(Value)

    def NormAmps(self, *args):
        """Normal Ampere rating"""
        # Getter
        if len(args) == 0:
            return self._lib.WireData_Get_NormAmps()

        # Setter
        (Value,) = args
        self._lib.WireData_Set_NormAmps(Value)

    def Rdc(self, *args):
        # Getter
        if len(args) == 0:
            return self._lib.WireData_Get_Rdc()

        # Setter
        (Value,) = args
        self._lib.WireData_Set_Rdc(Value)

    def Rac(self, *args):
        # Getter
        if len(args) == 0:
            return self._lib.WireData_Get_Rac()

        # Setter
        (Value,) = args
        self._lib.WireData_Set_Rac(Value)

    def GMRac(self, *args):
        # Getter
        if len(args) == 0:
            return self._lib.WireData_Get_GMRac()

        # Setter
        (Value,) = args
        self._lib.WireData_Set_GMRac(Value)

    def GMRUnits(self, *args):
        # Getter
        if len(args) == 0:
            return LineUnits(self._lib.WireData_Get_GMRUnits())

        # Setter
        (Value,) = args
        self._lib.WireData_Set_GMRUnits(Value)

    def Radius(self, *args):
        # Getter
        if len(args) == 0:
            return self._lib.WireData_Get_Radius()

        # Setter
        (Value,) = args
        self._lib.WireData_Set_Radius(Value)

    def RadiusUnits(self, *args):
        # Getter
        if len(args) == 0:
            return self._lib.WireData_Get_RadiusUnits()

        # Setter
        (Value,) = args
        self._lib.WireData_Set_RadiusUnits(Value)

    def ResistanceUnits(self, *args):
        # Getter
        if len(args) == 0:
            return LineUnits(self._lib.WireData_Get_ResistanceUnits())

        # Setter
        (Value,) = args
        self._lib.WireData_Set_ResistanceUnits(Value)

    def Diameter(self, *args):
        # Getter
        if len(args) == 0:
            return self._lib.WireData_Get_Diameter()

        # Setter
        (Value,) = args
        self._lib.WireData_Set_Diameter(Value)

    def CapRadius(self, *args):
        """Equivalent conductor radius for capacitance calcs. Specify this for bundled conductors. Defaults to same value as radius."""
        # Getter
        if len(args) == 0:
            return self._lib.WireData_Get_CapRadius()

        # Setter
        (Value,) = args
        self._lib.WireData_Set_CapRadius(Value)


_WireData = IWireData(api_util)

# For backwards compatibility, bind to the default instance
AllNames = _WireData.AllNames
CapRadius = _WireData.CapRadius
Count = _WireData.Count
Diameter = _WireData.Diameter
EmergAmps = _WireData.EmergAmps
First = _WireData.First
GMRUnits = _WireData.GMRUnits
GMRac = _WireData.GMRac
Idx = _WireData.Idx
Name = _WireData.Name
Next = _WireData.Next
NormAmps = _WireData.NormAmps
Rac = _WireData.Rac
Radius = _WireData.Radius
RadiusUnits = _WireData.RadiusUnits
Rdc = _WireData.Rdc
ResistanceUnits = _WireData.ResistanceUnits
_columns = _WireData._columns
__all__ = [
    "AllNames",
    "CapRadius",
    "Count",
    "Diameter",
    "EmergAmps",
    "First",
    "GMRUnits",
    "GMRac",
    "Idx",
    "Name",
    "Next",
    "NormAmps",
    "Rac",
    "Radius",
    "RadiusUnits",
    "Rdc",
    "ResistanceUnits",
]



from ._utils import api_util
from .Bases import Base


class IParser(Base):
    __slots__ = []

    __name__ = "Parser"
    _api_prefix = "Parser"
    _columns = [
        "Delimiters",
        "EndQuote",
        "CmdString",
        "BeginQuote",
        "WhiteSpace",
        "AutoIncrement",
    ]

    def Matrix(self, ExpectedOrder):
        """Use this property to parse a Matrix token in OpenDSS format.  Returns square matrix of order specified. Order same as default Fortran order: column by column."""
        return self._lib.Parser_Get_Matrix_GR(ExpectedOrder)

    def SymMatrix(self, ExpectedOrder):
        """Use this property to parse a matrix token specified in lower triangle form. Symmetry is forced."""
        return self._lib.Parser_Get_SymMatrix_GR(ExpectedOrder)

    def Vector(self, ExpectedSize):
        """Returns token as array of doubles. For parsing quoted array syntax."""
        return self._lib.Parser_Get_Vector_GR(ExpectedSize)

    def ResetDelimiters(self):
        """
        Reset the delimiters to their default values.

        Original COM help: https://opendss.epri.com/ResetDelimiters.html
        """
        self._lib.Parser_ResetDelimiters()

    def AutoIncrement(self, *args):
        """
        Default is FALSE. If TRUE, the parser automatically advances to next token after DblValue, IntValue, or StrValue. Simpler when you don't need to check for parameter names.

        Original COM help: https://opendss.epri.com/AutoIncrement.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Parser_Get_AutoIncrement()

        # Setter
        (Value,) = args
        self._lib.Parser_Set_AutoIncrement(Value)

    def BeginQuote(self, *args):
        """
        Get/Set String containing the the characters for Quoting in OpenDSS scripts. Matching pairs defined in EndQuote. Default is "'([{.

        Original COM help: https://opendss.epri.com/BeginQuote.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Parser_Get_BeginQuote()

        # Setter
        (Value,) = args
        self._lib.Parser_Set_BeginQuote(Value)

    def CmdString(self, *args):
        """
        String to be parsed. Loading this string resets the Parser to the beginning of the line. Then parse off the tokens in sequence.

        Original COM help: https://opendss.epri.com/CmdString.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Parser_Get_CmdString()

        # Setter
        (Value,) = args
        self._lib.Parser_Set_CmdString(Value)

    def DblValue(self):
        """
        Return next parameter as a double.

        Original COM help: https://opendss.epri.com/DblValue.html
        """
        return self._lib.Parser_Get_DblValue()

    def Delimiters(self, *args):
        """
        String defining hard delimiters used to separate token on the command string. Default is , and =. The = separates token name from token value. These override whitespace to separate tokens.

        Original COM help: https://opendss.epri.com/Delimiters.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Parser_Get_Delimiters()

        # Setter
        (Value,) = args
        self._lib.Parser_Set_Delimiters(Value)

    def EndQuote(self, *args):
        """
        String containing characters, in order, that match the beginning quote characters in BeginQuote. Default is "')]}

        Original COM help: https://opendss.epri.com/EndQuote.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Parser_Get_EndQuote()

        # Setter
        (Value,) = args
        self._lib.Parser_Set_EndQuote(Value)

    def IntValue(self):
        """
        Return next parameter as a long integer.

        Original COM help: https://opendss.epri.com/IntValue.html
        """
        return self._lib.Parser_Get_IntValue()

    def NextParam(self):
        """
        Get next token and return tag name (before = sign) if any. See AutoIncrement.

        Original COM help: https://opendss.epri.com/NextParam.html
        """
        return self._lib.Parser_Get_NextParam()

    def StrValue(self):
        """
        Return next parameter as a string

        Original COM help: https://opendss.epri.com/StrValue.html
        """
        return self._lib.Parser_Get_StrValue()

    def WhiteSpace(self, *args):
        """
        Get/set the characters used for White space in the command string.  Default is blank and Tab.

        Original COM help: https://opendss.epri.com/WhiteSpace.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Parser_Get_WhiteSpace()

        # Setter
        (Value,) = args
        self._lib.Parser_Set_WhiteSpace(Value)


_Parser = IParser(api_util)

# For backwards compatibility, bind to the default instance
AutoIncrement = _Parser.AutoIncrement
BeginQuote = _Parser.BeginQuote
CmdString = _Parser.CmdString
DblValue = _Parser.DblValue
Delimiters = _Parser.Delimiters
EndQuote = _Parser.EndQuote
IntValue = _Parser.IntValue
Matrix = _Parser.Matrix
NextParam = _Parser.NextParam
ResetDelimiters = _Parser.ResetDelimiters
StrValue = _Parser.StrValue
SymMatrix = _Parser.SymMatrix
Vector = _Parser.Vector
WhiteSpace = _Parser.WhiteSpace
_columns = _Parser._columns
__all__ = [
    "AutoIncrement",
    "BeginQuote",
    "CmdString",
    "DblValue",
    "Delimiters",
    "EndQuote",
    "IntValue",
    "Matrix",
    "NextParam",
    "ResetDelimiters",
    "StrValue",
    "SymMatrix",
    "Vector",
    "WhiteSpace",
]



from ._utils import api_util
from .Bases import Iterable
from dss_python_backend.enums import CapControlModes

class ICapControls(Iterable):
    __slots__ = []

    __name__ = "CapControls"
    _api_prefix = "CapControls"
    _columns = [
        "Name",
        "Idx",
        "Capacitor",
        "CTRatio",
        "PTRatio",
        "DeadTime",
        "Delay",
        "DelayOff",
        "Vmin",
        "Vmax",
        "UseVoltOverride",
        "Mode",
        "MonitoredObj",
        "MonitoredTerm",
        "OFFSetting",
        "ONSetting",
    ]

    def Reset(self):
        """
        Force a reset of this CapControl.

        Original COM help: https://opendss.epri.com/Reset.html
        """
        self._lib.CapControls_Reset()

    def CTRatio(self, *args):
        """
        Transducer ratio from primary current to control current.

        Original COM help: https://opendss.epri.com/CTratio.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_CTratio()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_CTratio(Value)

    def Capacitor(self, *args):
        """
        Name of the Capacitor that is controlled.

        Original COM help: https://opendss.epri.com/Capacitor.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_Capacitor()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_Capacitor(Value)

    def DeadTime(self, *args):
        """
        Dead time after capacitor is turned OFF before it can be turned back ON for the active CapControl.

        Default is 300 sec.

        Original COM help: https://opendss.epri.com/DeadTime.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_DeadTime()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_DeadTime(Value)

    def Delay(self, *args):
        """
        Time delay [s] to switch on after arming.  Control may reset before actually switching.

        Original COM help: https://opendss.epri.com/Delay.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_Delay()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_Delay(Value)

    def DelayOff(self, *args):
        """
        Time delay [s] before switching off a step. Control may reset before actually switching.

        Original COM help: https://opendss.epri.com/DelayOff.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_DelayOff()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_DelayOff(Value)

    def Mode(self, *args):
        """
        Type of automatic controller.

        Original COM help: https://opendss.epri.com/Mode.html
        """
        # Getter
        if len(args) == 0:
            return CapControlModes(self._lib.CapControls_Get_Mode())

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_Mode(Value)

    def MonitoredObj(self, *args):
        """
        Full name of the element that PT and CT are connected to.

        Original COM help: https://opendss.epri.com/MonitoredObj.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_MonitoredObj()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_MonitoredObj(Value)

    def MonitoredTerm(self, *args):
        """
        Terminal number on the element that PT and CT are connected to.

        Original COM help: https://opendss.epri.com/MonitoredTerm.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_MonitoredTerm()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_MonitoredTerm(Value)

    def OFFSetting(self, *args):
        """
        Threshold to switch off a step. See Mode for units.

        Original COM help: https://opendss.epri.com/OFFSetting.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_OFFSetting()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_OFFSetting(Value)

    def ONSetting(self, *args):
        """
        Threshold to arm or switch on a step.  See Mode for units.

        Original COM help: https://opendss.epri.com/ONSetting.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_ONSetting()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_ONSetting(Value)

    def PTRatio(self, *args):
        """
        Transducer ratio from primary feeder to control voltage.

        Original COM help: https://opendss.epri.com/PTratio.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_PTratio()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_PTratio(Value)

    def UseVoltOverride(self, *args):
        """
        Enables Vmin and Vmax to override the control Mode

        Original COM help: https://opendss.epri.com/UseVoltOverride.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_UseVoltOverride()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_UseVoltOverride(Value)

    def Vmax(self, *args):
        """
        With VoltOverride, swtich off whenever PT voltage exceeds this level.

        Original COM help: https://opendss.epri.com/Vmax.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_Vmax()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_Vmax(Value)

    def Vmin(self, *args):
        """
        With VoltOverride, switch ON whenever PT voltage drops below this level.

        Original COM help: https://opendss.epri.com/Vmin.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.CapControls_Get_Vmin()

        # Setter
        (Value,) = args
        self._lib.CapControls_Set_Vmin(Value)


_CapControls = ICapControls(api_util)

# For backwards compatibility, bind to the default instance
AllNames = _CapControls.AllNames
CTRatio = _CapControls.CTRatio
Capacitor = _CapControls.Capacitor
Count = _CapControls.Count
DeadTime = _CapControls.DeadTime
Delay = _CapControls.Delay
DelayOff = _CapControls.DelayOff
First = _CapControls.First
Idx = _CapControls.Idx
Mode = _CapControls.Mode
MonitoredObj = _CapControls.MonitoredObj
MonitoredTerm = _CapControls.MonitoredTerm
Name = _CapControls.Name
Next = _CapControls.Next
OFFSetting = _CapControls.OFFSetting
ONSetting = _CapControls.ONSetting
PTRatio = _CapControls.PTRatio
Reset = _CapControls.Reset
UseVoltOverride = _CapControls.UseVoltOverride
Vmax = _CapControls.Vmax
Vmin = _CapControls.Vmin
_columns = _CapControls._columns
__all__ = [
    "AllNames",
    "CTRatio",
    "Capacitor",
    "Count",
    "DeadTime",
    "Delay",
    "DelayOff",
    "First",
    "Idx",
    "Mode",
    "MonitoredObj",
    "MonitoredTerm",
    "Name",
    "Next",
    "OFFSetting",
    "ONSetting",
    "PTRatio",
    "Reset",
    "UseVoltOverride",
    "Vmax",
    "Vmin",
]



from ._utils import api_util
from .Bases import Base
from typing import List as List_

class IZIP(Base):
    __slots__ = []
    __name__ = "ZIP"
    _api_prefix = "ZIP"
    _columns = []

    """
    ZIP allows controlling the ZIP-compressed file functions from AltDSS/DSS C-API.

    It allows opening a ZIP file, and loading circuits directly from it, without requiring extracting the contents to files before reading.

    The implementation provides a specialization which allows more efficient access if the ZIP file is open and reused for many circuits. 
    Doing so reduces the overhead of the initial opening and indexing of the file contents.

    *Not available when using EPRI's OpenDSS distribution.*

    (**API Extension**)
    """

    def Open(self, FileName: str):
        """
        Opens and prepares a ZIP file to be used by the DSS text parser.
        Currently, the ZIP format support is limited by what is provided in the Free Pascal distribution.
        Besides that, the full filenames inside the ZIP must be shorter than 256 characters.
        The limitations should be removed in a future revision.

        **(API Extension)**
        """
        self._lib.ZIP_Open(FileName)

    def Close(self):
        """
        Closes the current open ZIP file

        **(API Extension)**
        """
        self._lib.ZIP_Close()

    def Redirect(self, FileInZip: str):
        """
        Runs a "Redirect" command inside the current (open) ZIP file.
        In the current implementation, all files required by the script must
        be present inside the ZIP, using relative paths. The only exceptions are
        memory-mapped files.

        **(API Extension)**
        """
        self._lib.ZIP_Redirect(FileInZip)

    def Extract(self, FileName: str) -> bytes:
        """
        Extracts the contents of the file "FileName" from the current (open) ZIP file.
        Returns a byte-string.

        **(API Extension)**
        """
        api_util = self._api_util
        if not isinstance(FileName, bytes):
            FileName = FileName.encode(api_util.codec)

        api_util.lib_unpatched.ZIP_Extract_GR(FileName)
        api_util._check_for_error()
        ptr, cnt = api_util.gr_int8_pointers
        return bytes(api_util.ffi.buffer(ptr[0], cnt[0]))


    def List(self, regexp: str="") -> List_[str]:
        """
        List of strings consisting of all names match the regular expression provided in regexp.
        If no expression is provided, all names in the current open ZIP are returned.

        See https://regex.sorokin.engineer/en/latest/regular_expressions.html for information on
        the expression syntax and options.

        **(API Extension)**
        """
        if regexp is None or not regexp:
            regexp = b""

        return self._lib.ZIP_List(regexp)

    def Contains(self, Name: str) -> bool:
        """
        Check if the given path name is present in the current ZIP file.

        **(API Extension)**
        """
        return self._lib.ZIP_Contains(Name)


_ZIP = IZIP(api_util)

# For backwards compatibility, bind to the default instance
Close = _ZIP.Close
Contains = _ZIP.Contains
Extract = _ZIP.Extract
List = _ZIP.List
Open = _ZIP.Open
Redirect = _ZIP.Redirect
_columns = _ZIP._columns
__all__ = ["Close", "Contains", "Extract", "List", "Open", "Redirect"]



"""
This package implements the OpenDSSDirect.py API using DSS C-API instead
of the official OpenDSS Direct DLL, sharing the CFFI backend and some other
features (plotting) with DSS-Python and AltDSS.
"""
try:
    # Version is now populated by the build script based
    # on the git tags. See pyproject.toml for more info.
    from ._version import __version__
except:
    __version__ = "0.0dev"

from . import utils
from . import enums
from .OpenDSSDirect import (
    dss,
    run_command,
    ActiveClass,
    Basic,
    Bus,
    CapControls,
    Capacitors,
    Circuit,
    CktElement,
    CtrlQueue,
    DSSCore,
    Element,
    Error,
    Executive,
    Fuses,
    Generators,
    Isource,
    LineCodes,
    Lines,
    LoadShape,
    Loads,
    Meters,
    Monitors,
    PDElements,
    PVsystems,
    Parallel,
    Parser,
    Progress,
    Properties,
    Reclosers,
    RegControls,
    Relays,
    Sensors,
    Settings,
    Solution,
    Storages,
    SwtControls,
    Text,
    Topology,
    Transformers,
    Vsources,
    XYCurves,
    YMatrix,
    ZIP,
    to_dss_python,
    to_altdss,
    NewContext,
    dss_lib,
    DSSException,
)
# Since run_command is deprecated, use these instead
Command = dss.Command
Commands = dss.Commands



from ._utils import api_util
from .Bases import Base


class IElement(Base):
    __slots__ = []

    __name__ = "Element"
    _api_prefix = "DSSElement"
    _columns = ["AllPropertyNames", "Name", "NumProperties"]

    def AllPropertyNames(self):
        """
        Array of strings containing the names of all properties for the active DSS object.

        Original COM help: https://opendss.epri.com/AllPropertyNames1.html
        """
        return self._lib.DSSElement_Get_AllPropertyNames()

    def Name(self):
        """
        Full Name of Active DSS Object (general element or circuit element).

        Original COM help: https://opendss.epri.com/Name5.html
        """
        return self._lib.DSSElement_Get_Name()

    def NumProperties(self):
        """
        Number of Properties for the active DSS object.

        Original COM help: https://opendss.epri.com/NumProperties1.html
        """
        return self._lib.DSSElement_Get_NumProperties()

    def ToJSON(self, options=0):
        """
        Returns the properties of the active DSS object as a JSON-encoded string.

        The `options` parameter contains bit-flags to toggle specific features.
        See `Obj_ToJSON` (C-API) for more, or `DSSObj.to_json` in Python.

        **(API Extension)**
        """
        return self._lib.DSSElement_ToJSON(options)

    def to_altdss(self):
        """
        Returns a Python object for the current active DSS object in this interface.

        Requires AltDSS-Python.

        *Available only for the AltDSS engine.*

        **(API Extension)**
        """
        ptr = self._api_util._lib.DSSElement_Get_Pointer()
        return self._api_util.get_dss_obj(ptr)


_Element = IElement(api_util)

# For backwards compatibility, bind to the default instance
AllPropertyNames = _Element.AllPropertyNames
Name = _Element.Name
NumProperties = _Element.NumProperties
ToJSON = _Element.ToJSON
to_altdss = _Element.to_altdss
_columns = _Element._columns
__all__ = ["AllPropertyNames", "Name", "NumProperties", "ToJSON", "to_altdss"]



from ._utils import api_util
from .Bases import Iterable


class IReactors(Iterable):
    """
    Reactor objects
    
    API Status: **(API Extension)** before 2024-10-04. Since then, functions partially marked as extensions (see each property/function documentation).
    """

    __name__ = "Reactors"
    _api_prefix = "Reactors"
    __slots__ = []
    _columns = [
        "Name",
        "Idx",
        "Phases",
        "Bus1",
        "Bus2",
        "SpecType",
        "kV",
        "kvar",
        "IsDelta",
        "Parallel",
        "LCurve",
        "RCurve",
        "R",
        "Rp",
        "X",
        "Z0",
        "Z1",
        "Z2",
        "Z",
        "Rmatrix",
        "Xmatrix",
    ]

    def SpecType(self):
        """
        How the reactor data was provided: 1=kvar, 2=R+jX, 3=R and X matrices, 4=sym components.
        Depending on this value, only some properties are filled or make sense in the context.

        Typically not used by end users.

        **(API Extension)**
        """
        return self._lib.Reactors_Get_SpecType()

    def IsDelta(self, *args):
        """
        Delta connection or wye?

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_IsDelta()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_IsDelta(Value)

    def Parallel(self, *args):
        """Indicates whether Rmatrix and Xmatrix are to be considered in parallel."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Parallel()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_Parallel(Value)

    def LmH(self, *args):
        """Inductance, mH. Alternate way to define the reactance, X, property."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_LmH()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_LmH(Value)

    def kV(self, *args):
        """For 2, 3-phase, kV phase-phase. Otherwise specify actual coil rating."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_kV()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_kV(Value)

    def kvar(self, *args):
        """Total kvar, all phases.  Evenly divided among phases. Only determines X. Specify R separately"""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_kvar()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_kvar(Value)

    def Phases(self, *args):
        """
        Number of phases.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Phases()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_Phases(Value)

    def Bus1(self, *args):
        """
        Name of first bus.
        Bus2 property will default to this bus, node 0, unless previously specified.
        Only Bus1 need be specified for a Yg shunt reactor.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Bus1()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_Bus1(Value)

    def Bus2(self, *args):
        """
        Name of 2nd bus. Defaults to all phases connected to first bus, node 0, (Shunt Wye Connection) except when Bus2 is specifically defined.
        Not necessary to specify for delta (LL) connection.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Bus2()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_Bus2(Value)

    def LCurve(self, *args):
        """Name of XYCurve object, previously defined, describing per-unit variation of phase inductance, L=X/w, vs. frequency. Applies to reactance specified by X, LmH, Z, or kvar property. L generally decreases somewhat with frequency above the base frequency, approaching a limit at a few kHz."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_LCurve()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_LCurve(Value)

    def RCurve(self, *args):
        """Name of XYCurve object, previously defined, describing per-unit variation of phase resistance, R, vs. frequency. Applies to resistance specified by R or Z property. If actual values are not known, R often increases by approximately the square root of frequency."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_RCurve()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_RCurve(Value)

    def R(self, *args):
        """Resistance (in series with reactance), each phase, ohms. This property applies to REACTOR specified by either kvar or X. See also help on Z."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_R()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_R(Value)

    def X(self, *args):
        """Reactance, each phase, ohms at base frequency. See also help on Z and LmH properties."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_X()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_X(Value)

    def Rp(self, *args):
        """Resistance in parallel with R and X (the entire branch). Assumed infinite if not specified."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Rp()

        # Setter
        (Value,) = args
        self._lib.Reactors_Set_Rp(Value)

    def Rmatrix(self, *args):
        """Resistance matrix, ohms at base frequency. Order of the matrix is the number of phases. Mutually exclusive to specifying parameters by kvar or X."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Rmatrix_GR()

        # Setter
        (Value,) = args
        Value, ValuePtr, ValueCount = self._prepare_float64_array(Value)
        self._lib.Reactors_Set_Rmatrix(ValuePtr, ValueCount)

    def Xmatrix(self, *args):
        """Reactance matrix, ohms at base frequency. Order of the matrix is the number of phases. Mutually exclusive to specifying parameters by kvar or X."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Xmatrix_GR()

        # Setter
        (Value,) = args
        Value, ValuePtr, ValueCount = self._prepare_float64_array(Value)
        self._lib.Reactors_Set_Xmatrix(ValuePtr, ValueCount)

    def Z(self, *args):
        """Alternative way of defining R and X properties. Enter a 2-element array representing R +jX in ohms."""
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Z_GR()

        # Setter
        (Value,) = args
        Value, ValuePtr, ValueCount = self._prepare_complex128_simple(Value)
        self._lib.Reactors_Set_Z(ValuePtr, ValueCount)

    def Z1(self, *args):
        """
        Positive-sequence impedance, ohms, as a 2-element array representing a complex number.

        If defined, Z1, Z2, and Z0 are used to define the impedance matrix of the REACTOR.

        Z1 MUST BE DEFINED TO USE THIS OPTION FOR DEFINING THE MATRIX.

        Side Effect: Sets Z2 and Z0 to same values unless they were previously defined.
        """
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Z1_GR()

        # Setter
        (Value,) = args
        Value, ValuePtr, ValueCount = self._prepare_complex128_simple(Value)
        self._lib.Reactors_Set_Z1(ValuePtr, ValueCount)

    def Z2(self, *args):
        """
        Negative-sequence impedance, ohms, as a 2-element array representing a complex number.

        Used to define the impedance matrix of the REACTOR if Z1 is also specified.

        Note: Z2 defaults to Z1 if it is not specifically defined. If Z2 is not equal to Z1, the impedance matrix is asymmetrical.
        """
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Z2_GR()

        # Setter
        (Value,) = args
        Value, ValuePtr, ValueCount = self._prepare_complex128_simple(Value)
        self._lib.Reactors_Set_Z2(ValuePtr, ValueCount)

    def Z0(self, *args):
        """
        Zero-sequence impedance, ohms, as a 2-element array representing a complex number.

        Used to define the impedance matrix of the REACTOR if Z1 is also specified.

        Note: Z0 defaults to Z1 if it is not specifically defined.
        """
        # Getter
        if len(args) == 0:
            return self._lib.Reactors_Get_Z0_GR()

        # Setter
        (Value,) = args
        Value, ValuePtr, ValueCount = self._prepare_complex128_simple(Value)
        self._lib.Reactors_Set_Z0(ValuePtr, ValueCount)


_Reactors = IReactors(api_util)

# For backwards compatibility, bind to the default instance
AllNames = _Reactors.AllNames
Bus1 = _Reactors.Bus1
Bus2 = _Reactors.Bus2
Count = _Reactors.Count
First = _Reactors.First
Idx = _Reactors.Idx
IsDelta = _Reactors.IsDelta
LCurve = _Reactors.LCurve
LmH = _Reactors.LmH
Name = _Reactors.Name
Next = _Reactors.Next
Parallel = _Reactors.Parallel
Phases = _Reactors.Phases
R = _Reactors.R
RCurve = _Reactors.RCurve
Rmatrix = _Reactors.Rmatrix
Rp = _Reactors.Rp
SpecType = _Reactors.SpecType
X = _Reactors.X
Xmatrix = _Reactors.Xmatrix
Z = _Reactors.Z
Z0 = _Reactors.Z0
Z1 = _Reactors.Z1
Z2 = _Reactors.Z2
kV = _Reactors.kV
kvar = _Reactors.kvar
_columns = _Reactors._columns
__all__ = [
    "AllNames",
    "Bus1",
    "Bus2",
    "Count",
    "First",
    "Idx",
    "IsDelta",
    "LCurve",
    "LmH",
    "Name",
    "Next",
    "Parallel",
    "Phases",
    "R",
    "RCurve",
    "Rmatrix",
    "Rp",
    "SpecType",
    "X",
    "Xmatrix",
    "Z",
    "Z0",
    "Z1",
    "Z2",
    "kV",
    "kvar",
]



import warnings
from ._utils import api_util, dss_py
from .Bases import Base
try:
    from ._version import __version__
except:
    __version__ = '0.0dev'

# from dss import DSSCompatFlags


class IBasic(Base):
    __name__ = "Basic"
    _api_prefix = "DSS"
    _columns = [
        "Version",
        "Classes",
        "NumUserClasses",
        "DataPath",
        "NumClasses",
        "NumCircuits",
        "UserClasses",
        "DefaultEditor",
    ]
    __slots__ = []

    def ClearAll(self):
        self._lib.DSS_ClearAll()

    def Reset(self):
        """
        This is a no-op function, does nothing. Left for compatibility.

        Original COM help: https://opendss.epri.com/Reset1.html
        """
        self._lib.DSS_Reset()

    def SetActiveClass(self, ClassName):
        return self._lib.DSS_SetActiveClass(ClassName)

    def Start(self, code):
        """
        This is a no-op function, does nothing. Left for compatibility.

        Calling `Start` in AltDSS/DSS-Extensions is required but that is already
        handled automatically, so the users do not need to call it manually,
        unless using AltDSS/DSS C-API directly without further tools.

        On EPRI's OpenDSS, `Start` also does nothing at all in the current
        Delphi versions. It is required for OpenDSS-C, but also handled behind
        the scenes on DSS-Extensions.

        Original COM help: https://opendss.epri.com/Start.html
        """
        return self._lib.DSS_Start(code)

    def Classes(self):
        """
        List of DSS intrinsic classes (names of the classes)

        Original COM help: https://opendss.epri.com/Classes1.html
        """
        return self._lib.DSS_Get_Classes()

    def DataPath(self, *args):
        """
        DSS Data File Path.  Default path for reports, etc. from DSS

        Original COM help: https://opendss.epri.com/DataPath.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_DataPath()

        # Setter
        (Value,) = args
        self._lib.DSS_Set_DataPath(Value)

    def DefaultEditor(self):
        """
        Returns the path name for the default text editor.

        Original COM help: https://opendss.epri.com/DefaultEditor.html
        """
        return self._lib.DSS_Get_DefaultEditor()

    def NumCircuits(self):
        """
        Number of Circuits currently defined

        Original COM help: https://opendss.epri.com/NumCircuits.html
        """
        return self._lib.DSS_Get_NumCircuits()

    def NumClasses(self):
        """
        Number of DSS intrinsic classes

        Original COM help: https://opendss.epri.com/NumClasses.html
        """
        return self._lib.DSS_Get_NumClasses()

    def NumUserClasses(self):
        """
        Number of user-defined classes

        Original COM help: https://opendss.epri.com/NumUserClasses.html
        """
        return self._lib.DSS_Get_NumUserClasses()

    def UserClasses(self):
        """
        List of user-defined classes

        Original COM help: https://opendss.epri.com/UserClasses.html
        """
        return self._lib.DSS_Get_UserClasses()

    def Version(self):
        """
        Get version string for the DSS.

        Original COM help: https://opendss.epri.com/Version.html
        """
        return dss_py.DSS.Version + f"\nOpenDSSDirect.py version: {__version__}"

    def AllowForms(self, *args):
        """
        Indicates whether text output is allowed or forms are used.  Disable to silence most output.

        Currently, forms/windows are only used for EPRI's OpenDSS distribution on Windows.

        Original COM help: https://opendss.epri.com/AllowForms.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_AllowForms()

        # Setter
        (value,) = args
        self._lib.DSS_Set_AllowForms(value)

    def AllowEditor(self, *args):
        """
        Gets/sets whether running the external editor for "Show" is allowed

        AllowEditor controls whether the external editor is used in commands like "Show".
        If you set to 0 (false), the editor is not executed. Note that other side effects,
        such as the creation of files, are not affected.

        **Deprecated:** Use `Settings.AllowEditor()` instead (same behavior, the setting was just moved there for better organization).

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            warnings.warn(
                '"AllowEditor" was moved to the Settings interface. This method still works, but will be removed in a future release. Please use `...Settings.AllowEditor` instead.',
                DeprecationWarning,
                stacklevel=2,
            )
            return self._lib.DSS_Get_AllowEditor()

        # Setter
        (value,) = args
        self._lib.DSS_Set_AllowEditor(value)

    def ShowPanel(self):
        if self._api_util._is_oddie:
            self._lib.Text_Set_Command("panel")

        return 0

    def NewCircuit(self, name: str):
        """
        Make a new circuit.

        Original COM help: https://opendss.epri.com/NewCircuit.html

        **Deprecated**: Use `OpenDSSDirect.NewCircuit()` directly (`NewCircuit()` in the parent object).
        """
        warnings.warn(
            '"NewCircuit" was moved to the parent object in the OpenDSSDirect class. This method still works, but will be removed in a future release. Please use `...AllowEditor` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        self._lib.DSS_NewCircuit(name)
        return "New Circuit" # self.ActiveCircuit

    def LegacyModels(self, *args):
        """
        LegacyModels was a flag used to toggle legacy (pre-2019) models for PVSystem, InvControl, Storage and
        StorageControl.
        In EPRI's OpenDSS version 9.0, the old models were removed. They were temporarily present here
        but were also removed in DSS C-API v0.13.0.

        **NOTE**: this property will be removed for v1.0. It is left to avoid breaking the current API too soon.

        **(API Extension)**
        """
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_LegacyModels()

        # Setter
        (Value,) = args
        self._lib.DSS_Set_LegacyModels(Value)

    def AllowChangeDir(self, *args):
        """
        If disabled, the engine will not change the active working directory during execution. E.g. a "compile"
        command will not "chdir" to the file path.

        If you have issues with long paths, enabling this might help in some scenarios.

        Defaults to True (allow changes, backwards compatible) in the 0.10.x versions of DSS C-API.
        This might change to False in future versions.

        This can also be set through the environment variable DSS_CAPI_ALLOW_CHANGE_DIR. Set it to 0 to
        disallow changing the active working directory.

        **Deprecated:** Use `Settings.AllowChangeDir` instead (same behavior, the setting was just moved there for better organization).

        **(API Extension)**
        """
        warnings.warn(
            '"AllowChangeDir" was moved to the Settings interface. This method still works, but will be removed in a future release. Please use `...Settings.AllowChangeDir` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_AllowChangeDir()

        # Setter
        (Value,) = args
        self._lib.DSS_Set_AllowChangeDir(Value)

    def AllowDOScmd(self, *args):
        """
        If enabled, the `DOScmd` command is allowed. Otherwise, an error is reported if the user tries to use it.

        Defaults to False/0 (disabled state). Users should consider DOScmd deprecated on DSS-Extensions.

        This can also be set through the environment variable DSS_CAPI_ALLOW_DOSCMD. Setting it to 1 enables
        the command.

        **Deprecated:** Use `Settings.AllowDOScmd` instead (same behavior, the setting was just moved there for better organization).

        **(API Extension)**
        """
        warnings.warn(
            '"AllowDOScmd" was moved to the Settings interface. This method still works, but will be removed in a future release. Please use `...Settings.AllowDOScmd` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_AllowDOScmd()

        # Setter
        (Value,) = args
        self._lib.DSS_Set_AllowDOScmd(Value)

    def COMErrorResults(self, *args):
        """
        If enabled, in case of errors or empty arrays, the API returns arrays with values compatible with
        EPRI's OpenDSS COM interface.

        For example, consider the function `Loads_Get_ZIPV`. If there is no active circuit or active load element:

        - In the disabled state (`COMErrorResults(False)`), the function will return "[]", an array with 0 elements.
        - In the enabled state (`COMErrorResults(True)`), the function will return "[0.0]" instead. This should
        be compatible with the return value of EPRI's OpenDSS COM interface.

        Defaults to false (disabled state) in AltDSS since the v0.15.x series.

        This does not affect the results when using EPRI's OpenDSS distribution through Oddie.

        This can also be set through the environment variable `DSS_CAPI_COM_DEFAULTS`. Setting it to 1 enables
        the legacy/COM behavior. The value can be toggled through the API at any time.

        **Deprecated:** Use `Settings.COMErrorResults` instead (same behavior, the setting was just moved there for better organization).

        **(API Extension)**
        """
        warnings.warn(
            '"COMErrorResults" was moved to the Settings interface. This method still works, but will be removed in a future release. Please use `...Settings.COMErrorResults` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_COMErrorResults()

        # Setter
        (Value,) = args
        self._lib.DSS_Set_COMErrorResults(Value)

    def AdvancedTypes(self, *args):
        """
        When enabled, there are **two side-effects**:

        - **Per DSS Context:** Complex arrays and complex numbers can be returned and consumed by the Python API.
        - **Global effect:** The low-level API provides matrix dimensions when available (`EnableArrayDimensions` is enabled).

        As a result, for example, `OpenDSSDirect.CktElement.Yprim()` is returned as a complex matrix instead
        of a plain array.

        When disabled, the legacy plain arrays are used and complex numbers cannot be consumed by the Python API.

        *Defaults to **False** for backwards compatibility.*

        **Deprecated:** Use `Settings.AdvancedTypes` instead (same behavior, the setting was just moved there for better organization).

        **(API Extension)**
        """
        warnings.warn('"AdvancedTypes" was moved to the Settings interface. This method still works, but will be removed in a future release. Please use `...Settings.AdvancedTypes` instead.', DeprecationWarning, stacklevel=2)

        # Getter
        if len(args) == 0:
            return self._lib.advanced_types

        # Setter
        (Value,) = args
        self._lib.advanced_types = bool(Value)

    def CompatFlags(self, *args):
        """
        Controls some compatibility flags introduced to toggle some behavior from EPRI's OpenDSS.

        **THE FLAGS ARE GLOBAL, affecting all AltDSS engines in the process.**
        CompatFlags for Oddie-loaded instances (OpenDSS and OpenDSS-C engines) are handled by the Oddie code itself,
        so it is global for each Oddie library.

        These flags may change for each version of DSS C-API, but the same value will not be reused. That is,
        when we remove a compatibility flag, it will have no effect but will also not affect anything else
        besides raising an error if the user tries to toggle a flag that was available in a previous version.

        We expect to keep a very limited number of flags. Since the flags are more transient than the other
        options/flags, it was preferred to add this generic function instead of a separate function per
        flag.

        See the enumeration `DSSCompatFlags` for available flags, including description.

        **Deprecated:** Use `Settings.CompatFlags` instead (same behavior, the setting was just moved there for better organization).

        **(API Extension)**
        """
        warnings.warn(
            '"CompatFlags" was moved to the Settings interface. This method still works, but will be removed in a future release. Please use `...Settings.CompatFlags` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        # Getter
        if len(args) == 0:
            return self._lib.DSS_Get_CompatFlags()

        # Setter
        (Value,) = args
        self._lib.DSS_Set_CompatFlags(Value)


_Basic = IBasic(api_util)

# For backwards compatibility, bind to the default instance
AdvancedTypes = _Basic.AdvancedTypes
AllowChangeDir = _Basic.AllowChangeDir
AllowDOScmd = _Basic.AllowDOScmd
AllowEditor = _Basic.AllowEditor
AllowForms = _Basic.AllowForms
COMErrorResults = _Basic.COMErrorResults
Classes = _Basic.Classes
ClearAll = _Basic.ClearAll
CompatFlags = _Basic.CompatFlags
DataPath = _Basic.DataPath
DefaultEditor = _Basic.DefaultEditor
LegacyModels = _Basic.LegacyModels
NewCircuit = _Basic.NewCircuit
NumCircuits = _Basic.NumCircuits
NumClasses = _Basic.NumClasses
NumUserClasses = _Basic.NumUserClasses
Reset = _Basic.Reset
SetActiveClass = _Basic.SetActiveClass
ShowPanel = _Basic.ShowPanel
Start = _Basic.Start
UserClasses = _Basic.UserClasses
Version = _Basic.Version
_columns = _Basic._columns
__all__ = [
    "AdvancedTypes",
    "AllowChangeDir",
    "AllowDOScmd",
    "AllowEditor",
    "AllowForms",
    "COMErrorResults",
    "Classes",
    "ClearAll",
    "CompatFlags",
    "DataPath",
    "DefaultEditor",
    "LegacyModels",
    "NewCircuit",
    "NumCircuits",
    "NumClasses",
    "NumUserClasses",
    "Reset",
    "SetActiveClass",
    "ShowPanel",
    "Start",
    "UserClasses",
    "Version",
]



from ._utils import api_util
from .Bases import Base


class IPDElements(Base):
    """
    The PDElements interface allows accessing some common properties and 
    methods shared across power delivery elements in the DSS engine.

    Users can iterate on all PD elements directly through this interface, 
    or enable a PD element through a dedicated interface (e.g. use `Lines.Name`, `Transformers.First/Next`)
    and access the properties here.

    If you are new to OpenDSS/AltDSS and this classic interface, please read the following document
    for an overview of the "active element" paradigm used by COM and the classic APIs:
        
    https://dss-extensions.org/classic_api.html#the-active-paradigm
    """

    __slots__ = []

    __name__ = "PDElements"
    _api_prefix = "PDElements"
    _columns = [
        "Name",
        "AccumulatedL",
        "ParentPDElement",
        "FromTerminal",
        "IsShunt",
        "NumCustomers",
        "SectionID",
        "FaultRate",
        "RepairTime",
        "TotalMiles",
        "TotalCustomers",
        "PctPermanent",
        "Lambda",
    ]

    def AccumulatedL(self):
        """
        Accumulated failure rate for this branch on downline

        *Requires a previous call to `RelCalc` command*

        Original COM help: https://opendss.epri.com/AccumulatedL.html
        """
        return self._lib.PDElements_Get_AccumulatedL()

    def Count(self):
        """
        Number of PD elements (including disabled elements)

        Original COM help: https://opendss.epri.com/Count12.html
        """
        return self._lib.PDElements_Get_Count()

    def FaultRate(self, *args):
        """
        Get/Set Number of failures per year.
        For LINE elements: Number of failures per unit length per year.
        """
        # Getter
        if len(args) == 0:
            return self._lib.PDElements_Get_FaultRate()

        # Setter
        (Value,) = args
        self._lib.PDElements_Set_FaultRate(Value)

    def First(self):
        """
        (read-only) Set the first enabled PD element to be the active element.
        Returns 0 if none found.
        """
        return self._lib.PDElements_Get_First()

    def FromTerminal(self):
        """
        (read-only) Number of the terminal of active PD element that is on the "from"
        side. This is set after the meter zone is determined.

        *Requires an energy meter with an updated zone.*
        """
        return self._lib.PDElements_Get_FromTerminal()

    def IsShunt(self):
        """
        (read-only) Boolean indicating of PD element should be treated as a shunt
        element rather than a series element. Applies to Capacitor and Reactor
        elements in particular.
        """
        return self._lib.PDElements_Get_IsShunt()

    def Lambda(self):
        """
        Failure rate for this branch. Faults per year including length of line.

        *Requires a previous call to `RelCalc` command*

        Original COM help: https://opendss.epri.com/Lambda1.html
        """
        return self._lib.PDElements_Get_Lambda()

    def Name(self, *args):
        """
        Get/Set name of active PD Element. Returns null string if active element
        is not PDElement type.
        """
        # Getter
        if len(args) == 0:
            return self._lib.PDElements_Get_Name()

        # Setter
        (Value,) = args
        self._lib.PDElements_Set_Name(Value)

    def Next(self):
        """
        (read-only) Advance to the next PD element in the circuit. Enabled elements
        only. Returns 0 when no more elements.
        """
        return self._lib.PDElements_Get_Next()

    def NumCustomers(self):
        """
        Number of customers, this branch

        *Requires an energy meter with an updated zone.*

        Original COM help: https://opendss.epri.com/Numcustomers.html
        """
        return self._lib.PDElements_Get_Numcustomers()

    def ParentPDElement(self):
        """
        (read-only) Sets the parent PD element to be the active circuit element.
        Returns 0 if no more elements upline.

        *Requires an energy meter with an updated zone.*
        """
        return self._lib.PDElements_Get_ParentPDElement()

    def RepairTime(self, *args):
        """
        Average repair time for this element in hours

        Original COM help: https://opendss.epri.com/RepairTime.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.PDElements_Get_RepairTime()

        # Setter
        (Value,) = args
        self._lib.PDElements_Set_RepairTime(Value)

    def SectionID(self):
        """
        Integer ID of the feeder section that this PDElement branch is part of

        *Requires a previous call to `RelCalc` command*

        Original COM help: https://opendss.epri.com/SectionID1.html
        """
        return self._lib.PDElements_Get_SectionID()

    def TotalMiles(self):
        """
        Total miles of line from this element to the end of the zone. For recloser siting algorithm.

        *Requires a previous call to `RelCalc` command*

        Original COM help: https://opendss.epri.com/TotalMiles1.html
        """
        return self._lib.PDElements_Get_TotalMiles()

    def TotalCustomers(self):
        """
        Total number of customers from this branch to the end of the zone

        *Requires a circuit with an energy meter with an updated zone.*

        Original COM help: https://opendss.epri.com/TotalCustomers1.html
        """
        return self._lib.PDElements_Get_Totalcustomers()

    def PctPermanent(self, *args):
        """
        Get/Set percent of faults that are permanent (require repair). Otherwise, fault is assumed to be transient/temporary.

        Original COM help: https://opendss.epri.com/pctPermanent.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.PDElements_Get_pctPermanent()

        # Setter
        (Value,) = args
        self._lib.PDElements_Set_pctPermanent(Value)

    def AllNames(self):
        """
        Array of strings consisting of all PD element names.

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllNames()

    def AllMaxCurrents(self, AllNodes=False):
        """
        Array of doubles with the maximum current across the conductors, for each PD
        element.

        By default, only the *first terminal* is used for the maximum current, matching
        the behavior of the "export capacity" command. Pass `AllNodes=True` to
        force the analysis to all terminals.

        See also:
        https://sourceforge.net/p/electricdss/discussion/beginners/thread/da5b93ca/

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllMaxCurrents_GR(AllNodes)

    def AllPctNorm(self, AllNodes=False):
        """
        Array of doubles with the maximum current across the conductors as a percentage
        of the Normal Ampere Rating, for each PD element.

        By default, only the *first terminal* is used for the maximum current, matching
        the behavior of the "export capacity" command. Pass `AllNodes=True` to
        force the analysis to all terminals.

        See also:
        https://sourceforge.net/p/electricdss/discussion/beginners/thread/da5b93ca/

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllPctNorm_GR(AllNodes)

    def AllPctEmerg(self, AllNodes=False):
        """
        Array of doubles with the maximum current across the conductors as a percentage
        of the Emergency Ampere Rating, for each PD element.

        By default, only the *first terminal* is used for the maximum current, matching
        the behavior of the "export capacity" command. Pass `AllNodes=True` to
        force the analysis to all terminals.

        See also:
        https://sourceforge.net/p/electricdss/discussion/beginners/thread/da5b93ca/

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllPctEmerg_GR(AllNodes)

    def AllCurrents(self):
        """
        Complex array of currents for all conductors, all terminals, for each PD element.

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllCurrents_GR()

    def AllCurrentsMagAng(self):
        """
        Complex array (magnitude and angle format) of currents for all conductors, all terminals, for each PD element.

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllCurrentsMagAng_GR()

    def AllCplxSeqCurrents(self):
        """
        Complex double array of Sequence Currents for all conductors of all terminals, for each PD elements.

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllCplxSeqCurrents_GR()

    def AllSeqCurrents(self):
        """
        Double array of the symmetrical component currents (magnitudes only) into each 3-phase terminal, for each PD element.

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllSeqCurrents_GR()

    def AllPowers(self):
        """
        Complex array of powers into each conductor of each terminal, for each PD element.

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllPowers_GR()

    def AllSeqPowers(self):
        """
        Complex array of sequence powers into each 3-phase terminal, for each PD element

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllSeqPowers_GR()

    def AllNumPhases(self):
        """
        Integer array listing the number of phases of all PD elements

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllNumPhases_GR()

    def AllNumConductors(self):
        """
        Integer array listing the number of conductors of all PD elements

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllNumConductors_GR()

    def AllNumTerminals(self):
        """
        Integer array listing the number of terminals of all PD elements

        **(API Extension)**
        """
        return self._lib.PDElements_Get_AllNumTerminals_GR()


_PDElements = IPDElements(api_util)

# For backwards compatibility, bind to the default instance
AccumulatedL = _PDElements.AccumulatedL
AllCplxSeqCurrents = _PDElements.AllCplxSeqCurrents
AllCurrents = _PDElements.AllCurrents
AllCurrentsMagAng = _PDElements.AllCurrentsMagAng
AllMaxCurrents = _PDElements.AllMaxCurrents
AllNames = _PDElements.AllNames
AllNumConductors = _PDElements.AllNumConductors
AllNumPhases = _PDElements.AllNumPhases
AllNumTerminals = _PDElements.AllNumTerminals
AllPctEmerg = _PDElements.AllPctEmerg
AllPctNorm = _PDElements.AllPctNorm
AllPowers = _PDElements.AllPowers
AllSeqCurrents = _PDElements.AllSeqCurrents
AllSeqPowers = _PDElements.AllSeqPowers
Count = _PDElements.Count
FaultRate = _PDElements.FaultRate
First = _PDElements.First
FromTerminal = _PDElements.FromTerminal
IsShunt = _PDElements.IsShunt
Lambda = _PDElements.Lambda
Name = _PDElements.Name
Next = _PDElements.Next
NumCustomers = _PDElements.NumCustomers
ParentPDElement = _PDElements.ParentPDElement
PctPermanent = _PDElements.PctPermanent
RepairTime = _PDElements.RepairTime
SectionID = _PDElements.SectionID
TotalCustomers = _PDElements.TotalCustomers
TotalMiles = _PDElements.TotalMiles
_columns = _PDElements._columns
__all__ = [
    "AccumulatedL",
    "AllCplxSeqCurrents",
    "AllCurrents",
    "AllCurrentsMagAng",
    "AllMaxCurrents",
    "AllNames",
    "AllNumConductors",
    "AllNumPhases",
    "AllNumTerminals",
    "AllPctEmerg",
    "AllPctNorm",
    "AllPowers",
    "AllSeqCurrents",
    "AllSeqPowers",
    "Count",
    "FaultRate",
    "First",
    "FromTerminal",
    "IsShunt",
    "Lambda",
    "Name",
    "Next",
    "NumCustomers",
    "ParentPDElement",
    "PctPermanent",
    "RepairTime",
    "SectionID",
    "TotalCustomers",
    "TotalMiles",
]



from ._utils import api_util
from .Bases import Iterable
from dss import CoreType as TransformerCoreType

class ITransformers(Iterable):
    __slots__ = []

    __name__ = "Transformers"
    _api_prefix = "Transformers"
    _columns = [
        "Name",
        "Idx",
        "XfmrCode",
        "IsDelta",
        "CoreType",
        "NumWindings",
        "Wdg",
        "NumTaps",
        "MinTap",
        "MaxTap",
        "Tap",
        "kV",
        "kVA",
        "R",
        "Xhl",
        "Xht",
        "Xlt",
        "Rneut",
        "Xneut",
        "RdcOhms",
        "WdgCurrents",
        "WdgVoltages",
        "LossesByType",
    ]

    def IsDelta(self, *args):
        """
        Active Winding delta or wye connection?

        Original COM help: https://opendss.epri.com/IsDelta3.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_IsDelta()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_IsDelta(Value)

    def MaxTap(self, *args):
        """
        Active Winding maximum tap in per-unit.

        Original COM help: https://opendss.epri.com/MaxTap.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_MaxTap()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_MaxTap(Value)

    def MinTap(self, *args):
        """
        Active Winding minimum tap in per-unit.

        Original COM help: https://opendss.epri.com/MinTap.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_MinTap()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_MinTap(Value)

    def NumTaps(self, *args):
        """
        Active Winding number of tap steps between MinTap and MaxTap.

        Original COM help: https://opendss.epri.com/NumTaps.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_NumTaps()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_NumTaps(Value)

    def NumWindings(self, *args):
        """
        Number of windings on this transformer. Allocates memory; set or change this property first.

        Original COM help: https://opendss.epri.com/NumWindings.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_NumWindings()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_NumWindings(Value)

    def R(self, *args):
        """
        Active Winding resistance in %

        Original COM help: https://opendss.epri.com/R.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_R()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_R(Value)

    def Rneut(self, *args):
        """
        Active Winding neutral resistance [ohms] for wye connections. Set less than zero for ungrounded wye.

        Original COM help: https://opendss.epri.com/Rneut1.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_Rneut()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_Rneut(Value)

    def Tap(self, *args):
        """
        Active Winding tap in per-unit.

        Original COM help: https://opendss.epri.com/Tap.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_Tap()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_Tap(Value)

    def Wdg(self, *args):
        """
        Active Winding Number from 1..NumWindings. Update this before reading or setting a sequence of winding properties (R, Tap, kV, kVA, etc.)

        Original COM help: https://opendss.epri.com/Wdg.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_Wdg()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_Wdg(Value)

    def XfmrCode(self, *args):
        """
        Name of an XfrmCode that supplies electrical parameters for this Transformer.

        Original COM help: https://opendss.epri.com/XfmrCode1.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_XfmrCode()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_XfmrCode(Value)

    def Xhl(self, *args):
        """
        Percent reactance between windings 1 and 2, on winding 1 kVA base. Use for 2-winding or 3-winding transformers.

        Original COM help: https://opendss.epri.com/Xhl.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_Xhl()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_Xhl(Value)

    def Xht(self, *args):
        """
        Percent reactance between windings 1 and 3, on winding 1 kVA base.  Use for 3-winding transformers only.

        Original COM help: https://opendss.epri.com/Xht.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_Xht()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_Xht(Value)

    def Xlt(self, *args):
        """
        Percent reactance between windings 2 and 3, on winding 1 kVA base. Use for 3-winding transformers only.

        Original COM help: https://opendss.epri.com/Xlt.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_Xlt()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_Xlt(Value)

    def Xneut(self, *args):
        """
        Active Winding neutral reactance [ohms] for wye connections.

        Original COM help: https://opendss.epri.com/Xneut1.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_Xneut()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_Xneut(Value)

    def kV(self, *args):
        """
        Active Winding kV rating.  Phase-phase for 2 or 3 phases, actual winding kV for 1 phase transformer.

        Original COM help: https://opendss.epri.com/kV3.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_kV()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_kV(Value)

    def kVA(self, *args):
        """
        Active Winding kVA rating. On winding 1, this also determines normal and emergency current ratings for all windings.

        Original COM help: https://opendss.epri.com/kva1.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_kVA()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_kVA(Value)

    def WdgVoltages(self):
        """
        Complex array of voltages for active winding

        **WARNING:** If the transformer has open terminal(s), results may be wrong, i.e. avoid using this
        in those situations. For more information, see https://github.com/dss-extensions/dss-extensions/issues/24

        Original COM help: https://opendss.epri.com/WdgVoltages.html
        """
        return self._lib.Transformers_Get_WdgVoltages_GR()

    def WdgCurrents(self):
        """
        All Winding currents (ph1, wdg1, wdg2,... ph2, wdg1, wdg2 ...)

        **WARNING:** If the transformer has open terminal(s), results may be wrong, i.e. avoid using this
        in those situations. For more information, see https://github.com/dss-extensions/dss-extensions/issues/24

        Original COM help: https://opendss.epri.com/WdgCurrents.html
        """
        return self._lib.Transformers_Get_WdgCurrents_GR()

    def strWdgCurrents(self):
        """
        All winding currents in CSV string form like the WdgCurrents property

        **WARNING:** If the transformer has open terminal(s), results may be wrong, i.e. avoid using this
        in those situations. For more information, see https://github.com/dss-extensions/dss-extensions/issues/24
        """
        return self._lib.Transformers_Get_strWdgCurrents()

    def CoreType(self, *args):
        """
        Transformer Core Type: 0=Shell; 1=1ph; 3-3leg; 4=4-Leg; 5=5-leg; 9=Core-1-phase

        Original COM help: https://opendss.epri.com/CoreType.html
        """
        # Getter
        if len(args) == 0:
            return TransformerCoreType(self._lib.Transformers_Get_CoreType())

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_CoreType(Value)

    def RdcOhms(self, *args):
        """
        dc Resistance of active winding in ohms for GIC analysis

        Original COM help: https://opendss.epri.com/RdcOhms.html
        """
        # Getter
        if len(args) == 0:
            return self._lib.Transformers_Get_RdcOhms()

        # Setter
        (Value,) = args
        self._lib.Transformers_Set_RdcOhms(Value)

    def LossesByType(self):
        """
        Complex array with the losses by type (total losses, load losses, no-load losses), in VA, for the current active transformer

        **(API Extension)**
        """
        return self._lib.Transformers_Get_LossesByType_GR()

    def AllLossesByType(self):
        """
        Complex array with the losses by type (total losses, load losses, no-load losses), in VA, concatenated for ALL transformers

        **(API Extension)**
        """
        return self._lib.Transformers_Get_AllLossesByType_GR()


_Transformers = ITransformers(api_util)

# For backwards compatibility, bind to the default instance
AllLossesByType = _Transformers.AllLossesByType
AllNames = _Transformers.AllNames
CoreType = _Transformers.CoreType
Count = _Transformers.Count
First = _Transformers.First
Idx = _Transformers.Idx
IsDelta = _Transformers.IsDelta
LossesByType = _Transformers.LossesByType
MaxTap = _Transformers.MaxTap
MinTap = _Transformers.MinTap
Name = _Transformers.Name
Next = _Transformers.Next
NumTaps = _Transformers.NumTaps
NumWindings = _Transformers.NumWindings
R = _Transformers.R
RdcOhms = _Transformers.RdcOhms
Rneut = _Transformers.Rneut
Tap = _Transformers.Tap
Wdg = _Transformers.Wdg
WdgCurrents = _Transformers.WdgCurrents
WdgVoltages = _Transformers.WdgVoltages
XfmrCode = _Transformers.XfmrCode
Xhl = _Transformers.Xhl
Xht = _Transformers.Xht
Xlt = _Transformers.Xlt
Xneut = _Transformers.Xneut
kV = _Transformers.kV
kVA = _Transformers.kVA
strWdgCurrents = _Transformers.strWdgCurrents
_columns = _Transformers._columns
__all__ = [
    "AllLossesByType",
    "AllNames",
    "CoreType",
    "Count",
    "First",
    "Idx",
    "IsDelta",
    "LossesByType",
    "MaxTap",
    "MinTap",
    "Name",
    "Next",
    "NumTaps",
    "NumWindings",
    "R",
    "RdcOhms",
    "Rneut",
    "Tap",
    "Wdg",
    "WdgCurrents",
    "WdgVoltages",
    "XfmrCode",
    "Xhl",
    "Xht",
    "Xlt",
    "Xneut",
    "kV",
    "kVA",
    "strWdgCurrents",
]
